from __future__ import annotations

import typing

import toolcli
import toolstr

import ctc.config
from ctc import cli
from ctc import evm
from ctc import spec
from ctc.cli import cli_utils


command_help = """output ERC20 balances of blocks / addresses / tokens"""


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_balances_command,
        'help': command_help,
        'args': [
            {'name': 'args', 'nargs': '+', 'help': '<see above>'},
            {'name': '--block', 'help': 'block number'},
            {
                'name': '--wallets',
                'nargs': '+',
                'help': 'wallets to get balances of',
            },
            {
                'name': '--blocks',
                'nargs': '+',
                'help': 'block numbers to get balances at',
            },
            {
                'name': '--erc20s',
                'nargs': '+',
                'help': 'ERC20 addresses to get balances of',
            },
            #
            {
                'name': '--raw',
                'action': 'store_true',
                'help': 'whether to skip normalizing by ERC20 decimals',
            },
            {
                'name': '--export',
                'default': 'stdout',
                'help': 'file path for output (.json or .csv)',
            },
            {
                'name': '--overwrite',
                'action': 'store_true',
                'help': 'specify that output path can be overwritten',
            },
            {'name': '-n', 'help': 'show n data points', 'type': int},
        ],
        'examples': {
            'WALLET --erc20s ERC20S [--block BLOCK]': {
                'description': 'version 1: balance of single wallet across multiple tokens',
                'runnable': False,
            },
            'ERC20 [--block BLOCK] [--wallets WALLETS]': {
                'description': 'version 2: balances of multiple wallets in single block (default = all wallets)',
                'runnable': False,
            },
            'ERC20 WALLET --blocks BLOCKS': {
                'description': 'version 3: balance of single wallet across multiple blocks',
                'runnable': False,
            },
        },
    }


async def async_balances_command(
    *,
    args: typing.Sequence[str],
    block: typing.Optional[spec.BlockNumberReference],
    wallets: typing.Optional[typing.Sequence[str]],
    blocks: typing.Optional[typing.Sequence[str]],
    erc20s: typing.Optional[typing.Sequence[str]],
    raw: bool,
    export: str,
    overwrite: bool,
    n: int | None,
) -> None:

    import polars as pl

    if wallets is not None:
        wallets = await evm.async_resolve_addresses(wallets, block=block)
    if erc20s is not None:
        erc20s = await evm.async_resolve_addresses(erc20s, block=block)

    indent = None

    if len(args) == 1 and erc20s is not None:
        # multiple erc20s, single wallet, single block
        if wallets is not None or blocks is not None:
            raise Exception(
                'can only specify one of --erc20s --wallets, --blocks'
            )
        wallet = args[0]

        if block is None:
            block = 'latest'
        block = await evm.async_block_number_to_int(block)
        wallet = await evm.async_resolve_address(wallet, block=block)
        symbols_coroutine = evm.async_get_erc20s_symbols(erc20s)
        balances = await evm.async_get_erc20s_balances(
            wallet=wallet,
            tokens=erc20s,
            block=block,
            normalize=(not raw),
        )
        symbols = await symbols_coroutine
        data = {'balance': balances, 'symbol': symbols, 'erc20_address': erc20s}
        df = pl.DataFrame(data)
        output_data: typing.Union[spec.DataFrame, spec.Series] = df

        toolstr.print_text_box('ERC20 balances in wallet')
        print('- wallet:', wallet)
        print('- block:', block)
        print('- n_tokens:', len(erc20s))
        print()
        print()
        indent = '    '

    elif len(args) == 1:
        # single erc20, multiple wallets, single block
        if blocks is not None or erc20s is not None:
            raise Exception(
                'can only specify one of --erc20s --wallets, --blocks'
            )
        erc20 = args[0]

        if block is None:
            block = 'latest'
        block = await evm.async_block_number_to_int(block)
        erc20 = await evm.async_resolve_address(erc20, block=block)
        symbol_coroutine = evm.async_get_erc20_symbol(erc20)

        if wallets is not None:
            balances = await evm.async_get_erc20_balances_of_addresses(
                wallets=wallets,
                token=erc20,
                block=block,
                normalize=(not raw),
            )
            symbol = await symbol_coroutine

            df = pl.DataFrame({'balance': balances, 'address': wallets})
            output_data = df

            print()
            toolstr.print_text_box(symbol + ' Balances')
            print('- token:', erc20)
            print('- symbol:', symbol)
            print('- block:', block)
            print()
            print()

        else:

            import eth_abi_lite

            try:
                symbol = await symbol_coroutine
            except eth_abi_lite.exceptions.InsufficientDataBytes:
                print(
                    'this is an EOA, see `ctc erc20 balances -h` for proper syntax'
                )
                return

            transfers = await evm.async_get_erc20_transfers(
                erc20,
                start_block=None,
                end_block=block,
                normalize=False,
            )
            df = await evm.async_get_erc20_balances_from_transfers(
                transfers=transfers, normalize=(not raw)
            )
            output_data = df

            if n is None:
                n = 20

            print()
            toolstr.print_text_box(symbol + ' Balances')
            print('- token:', erc20)
            print('- symbol:', symbol)
            print('- block:', block)
            print()
            print()
            toolstr.print_header('Showing Top ' + str(n) + ' Holders')
            print()
            indent = '    '

    elif len(args) == 2 and blocks is not None:
        # single erc20, single address, multiple blocks
        if erc20s is not None or wallets is not None:
            raise Exception(
                'can only specify one of --erc20s --wallets, --blocks'
            )
        erc20, wallet = args
        resolved_blocks = await cli_utils.async_parse_block_slice(blocks, n=n)

        timestamps = await evm.async_get_block_timestamps(resolved_blocks)

        erc20 = await evm.async_resolve_address(
            erc20, block=resolved_blocks[-1]
        )
        wallet = await evm.async_resolve_address(
            wallet, block=resolved_blocks[-1]
        )
        balances = await evm.async_get_erc20_balance_by_block(
            wallet=wallet,
            token=erc20,
            blocks=resolved_blocks,
            normalize=(not raw),
            empty_token=0,
        )

        if export == 'stdout':
            styles = cli.get_cli_styles()

            # print header
            toolstr.print_text_box(
                'Historical ERC20 Balances', style=styles['title']
            )
            print()
            toolstr.print_table(
                rows=[['token', erc20], ['wallet', wallet]],
                column_justify=['right', 'left'],
                border=styles['comment'],
                column_styles=[styles['option'], styles['metavar']],
            )

            import tooltime

            # print balance history table
            print()
            print()
            rows = []
            for block, timestamp, balance in zip(resolved_blocks, timestamps, balances):
                age = tooltime.get_age(timestamp, 'TimelengthPhrase')
                age = ', '.join(age.split(', ')[:1])
                row = [tooltime.timestamp_to_iso(timestamp), age, block, balance]
                rows.append(row)
            labels = ['timestamp', 'age', 'block', 'balance']
            if n is None:
                n = 21
            toolstr.print_table(
                rows,
                labels=labels,
                border=styles['comment'],
                label_style=styles['title'],
                column_formats={'balance': {'order_of_magnitude': True, 'trailing_zeros': True}},
                column_styles={
                    'balance': styles['description'] + ' bold',
                },
                # indent=4,
                limit_rows=n,
            )

            # print balance history chart
            def formatter(xval: typing.Any) -> str:
                return toolstr.format(round(xval))

            xvals = resolved_blocks
            yvals = balances
            plot = toolstr.render_line_plot(
                xvals=xvals,
                yvals=yvals,
                n_rows=10,
                n_columns=60,
                line_style=styles['description'],
                chrome_style=styles['comment'],
                tick_label_style=styles['metavar'],
                xaxis_kwargs={'formatter': formatter},
                char_dict=ctc.config.get_cli_chart_charset(),
            )
            print()
            print()
            toolstr.print(
                toolstr.hjustify('ETH balance over time', 'center', 70),
                indent=4,
                style=styles['title'],
            )
            toolstr.print(plot, indent=4)

            return

        else:
            df = pl.DataFrame({'balance': balances, 'block': resolved_blocks})
            output_data = df

    else:
        raise Exception('invalid inputs')

    cli_utils.output_data(
        output_data,
        output=export,
        overwrite=overwrite,
        top=n,
        indent=indent,
    )



# TODO: all functions should accept `pool` as a positional argument
from __future__ import annotations

import typing

from ctc import evm
from ctc import rpc
from ctc import spec
from ctc.toolbox import nested_utils

from . import balancer_spec
from . import pool_metadata


pool_function_abis: typing.Mapping[str, spec.FunctionABI] = {
    'getNormalizedWeights': {
        'inputs': [],
        'name': 'getNormalizedWeights',
        'outputs': [
            {
                'internalType': 'uint256[]',
                'name': '',
                'type': 'uint256[]',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'getSwapFeePercentage': {
        'inputs': [],
        'name': 'getSwapFeePercentage',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
}


#
# # weights
#


async def async_get_pool_weights_raw(
    pool_address: spec.ContractAddress,
    block: spec.BlockNumberReference = 'latest',
    *,
    context: spec.Context = None,
) -> typing.Union[typing.Sequence[int], typing.Sequence[float]]:

    result = await rpc.async_eth_call(
        to_address=pool_address,
        function_abi=pool_function_abis['getNormalizedWeights'],
        block_number=block,
        context=context,
    )
    if not isinstance(result, (tuple, list)) or not all(
        isinstance(item, (int, float)) for item in result
    ):
        raise Exception('invalid rpc result')
    return typing.cast(
        typing.Union[typing.Sequence[int], typing.Sequence[float]], result
    )


async def async_get_pool_weights(
    pool_address: spec.ContractAddress,
    block: spec.BlockNumberReference = 'latest',
    *,
    normalize: bool = True,
    context: spec.Context = None,
) -> typing.Union[
    dict[spec.ContractAddress, int],
    dict[spec.ContractAddress, float],
]:
    import asyncio

    tokens_coroutine = pool_metadata.async_get_pool_tokens(
        pool_address=pool_address, block=block, context=context
    )
    weights_coroutine = async_get_pool_weights_raw(
        pool_address=pool_address, block=block, context=context
    )

    tokens, weights = await asyncio.gather(tokens_coroutine, weights_coroutine)

    if normalize:
        weights = [weight / 1e18 for weight in weights]

    return dict(zip(tokens, weights))


async def async_get_pool_weights_by_block(
    pool_address: spec.ContractAddress,
    blocks: typing.Sequence[spec.BlockNumberReference],
    *,
    normalize: bool = True,
    context: spec.Context = None,
) -> typing.Union[
    dict[spec.BlockNumberReference, int],
    dict[spec.BlockNumberReference, float],
]:
    from ctc import config

    context = config.update_context(context, merge_provider={'chunk_size': 100})

    weights = await rpc.async_batch_eth_call(
        to_address=pool_address,
        function_abi=pool_function_abis['getNormalizedWeights'],
        block_numbers=blocks,
        context=context,
    )

    if normalize:
        weights = [
            [block_weight / 1e18 for block_weight in block_weights]
            for block_weights in weights
        ]

    return dict(zip(blocks, weights))


#
# # fees
#


async def async_get_pool_fees(
    pool_address: spec.ContractAddress,
    *,
    block: spec.BlockNumberReference = 'latest',
    normalize: bool = True,
    context: spec.Context = None,
) -> typing.Union[int, float]:

    fees = await rpc.async_eth_call(
        to_address=pool_address,
        function_abi=pool_function_abis['getSwapFeePercentage'],
        block_number=block,
        context=context,
    )

    if not isinstance(fees, int):
        raise Exception('invalid rpc result')
    fees_result: int | float = fees

    if normalize:
        fees_result = fees_result / 1e18

    return fees_result


#
# # balances
#


async def async_get_pool_balances(
    *,
    pool_address: typing.Optional[spec.ContractAddress] = None,
    pool_id: typing.Optional[spec.HexData] = None,
    block: spec.BlockNumberReference | None = None,
    vault: typing.Optional[spec.ContractAddress] = None,
    normalize: bool = True,
    context: spec.Context = None,
) -> typing.Union[dict[spec.Address, int], dict[spec.Address, float]]:

    if block is None:
        block = 'latest'

    if vault is None:
        vault = balancer_spec.vault
    if pool_id is None:
        if pool_address is None:
            raise Exception('must specify pool_id or pool_address')
        pool_id = await pool_metadata.async_get_pool_id(
            pool_address,
            block=block,
            context=context,
        )

    pool_tokens = await rpc.async_eth_call(
        to_address=vault,
        function_abi=balancer_spec.vault_function_abis['getPoolTokens'],
        function_parameters=[pool_id],
        block_number=block,
        package_named_outputs=True,
        context=context,
    )

    pool_balances = dict(zip(pool_tokens['tokens'], pool_tokens['balances']))

    if normalize:
        tokens = pool_balances.keys()
        decimals = await evm.async_get_erc20s_decimals(
            tokens=tokens,
            block=block,
            context=context,
        )
        for token, decimal in zip(tokens, decimals):
            pool_balances[token] /= 10**decimal

    return pool_balances


async def async_get_pool_balances_by_block(
    *,
    blocks: typing.Sequence[spec.BlockNumberReference],
    pool_address: typing.Optional[spec.ContractAddress] = None,
    pool_id: typing.Optional[spec.HexData] = None,
    vault: typing.Optional[spec.ContractAddress] = None,
    normalize: bool = True,
    context: spec.Context = None,
) -> typing.Union[dict[spec.Address, list[int | float]]]:
    import asyncio

    coroutines = [
        async_get_pool_balances(
            pool_address=pool_address,
            pool_id=pool_id,
            block=block,
            vault=vault,
            normalize=normalize,
            context=context,
        )
        for block in blocks
    ]

    balances_by_block = await asyncio.gather(*coroutines)

    return nested_utils.list_of_dicts_to_dict_of_lists(balances_by_block)



from __future__ import annotations

import asyncio
import typing

from ctc import spec
import ctc.rpc

if typing.TYPE_CHECKING:
    import polars as pl


token0_abi = {
    'inputs': [],
    'name': 'token0',
    'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
    'stateMutability': 'view',
    'type': 'function',
}

token1_abi = {
    'inputs': [],
    'name': 'token1',
    'outputs': [{'internalType': 'address', 'name': '', 'type': 'address'}],
    'stateMutability': 'view',
    'type': 'function',
}

fee_abi = {
    'inputs': [],
    'name': 'fee',
    'outputs': [{'internalType': 'uint24', 'name': '', 'type': 'uint24'}],
    'stateMutability': 'view',
    'type': 'function',
}


async def _async_get_pool_fee(
    pool: str, context: spec.Context | None
) -> int | None:
    try:
        return await ctc.rpc.async_eth_call(
            to_address=pool,
            function_abi=fee_abi,
            context=context,
        )
    except Exception:
        return None


async def async_get_pool_summaries(
    pools: typing.Sequence[str],
    context: spec.Context | None = None,
) -> pl.DataFrame:
    import polars as pl

    # get token0
    coroutines = [
        ctc.rpc.async_eth_call(
            to_address=contract_address,
            function_abi=token0_abi,
            context=context,
        )
        for contract_address in pools
    ]
    token0s = await asyncio.gather(*coroutines)

    # get token1
    coroutines = [
        ctc.rpc.async_eth_call(
            to_address=contract_address,
            function_abi=token1_abi,
            context=context,
        )
        for contract_address in pools
    ]
    token1s = await asyncio.gather(*coroutines)

    # get token0 symbols
    coroutines = [
        ctc.async_get_erc20_symbol(
            token0,
            convert_reverts_to='UNKNOWN',
            context=context,
        )
        for token0 in token0s
    ]
    token0_symbols = await asyncio.gather(*coroutines)

    # get token1 symbols
    coroutines = [
        ctc.async_get_erc20_symbol(
            token1,
            convert_reverts_to='UNKNOWN',
            context=context,
        )
        for token1 in token1s
    ]
    token1_symbols = await asyncio.gather(*coroutines)

    # get token0 decimals
    coroutines = [
        ctc.async_get_erc20_decimals(
            token0,
            convert_reverts_to_none=True,
            context=context,
        )
        for token0 in token0s
    ]
    token0_decimals = await asyncio.gather(*coroutines)

    # get token1 decimals
    coroutines = [
        ctc.async_get_erc20_decimals(
            token1,
            convert_reverts_to_none=True,
            context=context,
        )
        for token1 in token1s
    ]
    token1_decimals = await asyncio.gather(*coroutines)

    # get fees
    coroutines = [
        _async_get_pool_fee(contract_address, context=context)
        for contract_address in pools
    ]
    pool_fees = await asyncio.gather(*coroutines)

    # create DataFrame
    pool_summaries = pl.DataFrame(
        [
            pl.Series(pools).alias('pool'),
            pl.Series(token0_symbols).alias('token0_symbol'),
            pl.Series(token1_symbols).alias('token1_symbol'),
            pl.Series(token0_decimals).alias('token0_decimals'),
            pl.Series(token1_decimals).alias('token1_decimals'),
            pl.Series(token0s).alias('token0_address'),
            pl.Series(token1s).alias('token1_address'),
            pl.Series(pool_fees).alias('pool_fee'),
        ]
    )

    # add pool_name
    pool_summaries = pool_summaries.with_columns(
        (
            pl.col('token0_symbol')
            + '_'
            + pl.col('token1_symbol')
            + '_'
            + (pl.col('pool_fee') / 100).cast(int).cast(str)
        ).alias('pool_name')
    )

    return pool_summaries


async def async_get_swaps_pool_metadata(
    swaps: pl.DataFrame, context: spec.Context = None
) -> pl.DataFrame:
    pools = sorted(swaps['contract_address'].unique().to_list())
    pool_summaries = await async_get_pool_summaries(pools, context=context)
    joined = swaps.join(
        pool_summaries,
        left_on='contract_address',
        right_on='pool',
        how='left',
    )
    return joined[
        [column for column in pool_summaries.columns if column != 'pool']
    ]


default_token_categories = {
    'Stable': {
        '0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48': 'USDC',
        '0xdac17f958d2ee523a2206206994597c13d831ec7': 'USDT',
        '0x6b175474e89094c44da98b954eedeac495271d0f': 'DAI',
    },
    'ETH': {
        '0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2': 'WETH',
    },
    'BTC': {
        '0x2260fac5e5542a773aa44fbcfedf7c193bc2c599': 'WBTC',
    },
}

default_pair_categories = [
    ('Stable', 'Stable', 'Stable-Stable'),
    ('Stable', 'ETH', 'Stable-ETH'),
    ('ETH', 'Stable', 'Stable-ETH'),
    ('Stable', 'BTC', 'Stable-BTC'),
    ('BTC', 'Stable', 'Stable-BTC'),
    ('Stable', 'Other', 'Stable-Other'),
    ('Other', 'Stable', 'Stable-Other'),
    ('ETH', 'ETH', 'ETH-ETH'),
    ('ETH', 'BTC', 'ETH-BTC'),
    ('BTC', 'ETH', 'ETH-BTC'),
    ('ETH', 'Other', 'ETH-Other'),
    ('Other', 'ETH', 'ETH-Other'),
    ('BTC', 'BTC', 'BTC-BTC'),
    ('BTC', 'Other', 'BTC-Other'),
    ('Other', 'BTC', 'BTC-Other'),
    ('Other', 'Other', 'Other-Other'),
]


def get_swaps_pair_categories(
    swaps: spec.DataFrame,
    token_categories: typing.Mapping[str, typing.Mapping[str, str]]
    | None = None,
    pair_categories: typing.Sequence[tuple[str, str, str]] | None = None,
) -> spec.DataFrame:
    import polars as pl

    if token_categories is None:
        token_categories = default_token_categories
    if pair_categories is None:
        pair_categories = default_pair_categories

    df = pl.DataFrame(
        [
            [category, address]
            for category, tokens in token_categories.items()
            for address in tokens
        ],
        orient='row',
        schema=['category', 'address'],
    )

    cats = swaps.join(
        df.rename({'address': 'token0_address', 'category': 'token0_category'}),
        on='token0_address',
        how='left',
    ).join(
        df.rename({'address': 'token1_address', 'category': 'token1_category'}),
        on='token1_address',
        how='left',
    )
    cats = cats.select(
        pl.col('token0_category').fill_null('Other'),
        pl.col('token1_category').fill_null('Other'),
    )

    df = pl.DataFrame(
        pair_categories,
        orient='row',
        schema=['token0_category', 'token1_category', 'pair_category'],
    )

    return cats.join(df, on=['token0_category', 'token1_category'], how='left')




from __future__ import annotations

import toolcli
import toolsql

import ctc


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': login_command,
        'help': 'login to database with interactive shell',
        'examples': [''],
    }


def login_command() -> None:
    db_config = ctc.config.get_db_config('main')
    toolsql.login(db_config)




from __future__ import annotations

import typing

import toolcli
import tooltime
import toolstr

from ctc import cli
from ctc import evm
from ctc import rpc
from ctc import spec

from ctc.cli import cli_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_blocks_command,
        'help': 'output information about blocks',
        'args': [
            {'name': 'blocks', 'nargs': '+', 'help': 'block range to fetch'},
            {
                'name': '-n',
                'help': 'number of blocks in block slice',
                'type': int,
            },
            {
                'name': '--attributes',
                'nargs': '+',
                'help': 'attributes to fetch from each block',
            },
            {
                'name': '--export',
                'default': 'stdout',
                'help': 'file path for output (.json or .csv)',
            },
            {
                'name': '--overwrite',
                'action': 'store_true',
                'help': 'specify that output path can be overwritten',
            },
            {'name': '--provider', 'help': 'rpc provider to use'},
        ],
        'examples': {
            '14000000:14000010': 'get blocks 14000000 through 14000010',
            '14000000:15000000:100000': 'get blocks 14000000 through 15000000, every 100000 blocks',
            '14000000:15000000 -n 10': 'get blocks 14000000 through 15000000, every 100000 blocks',
            '14000000:14000100 --attributes timestamp,number': 'get blocks 14000000 through 14000100, with specific attributes',
        },
    }


async def async_blocks_command(
    *,
    blocks: typing.Sequence[str],
    n: int | None,
    attributes: typing.Optional[typing.Sequence[str]],
    export: str,
    overwrite: bool,
    provider: typing.Optional[str],
) -> None:
    import polars as pl

    if attributes is None:
        # gas stats as well
        attributes = [
            'number',
            'timestamp',
            'age',
            'median_gas',
        ]
    elif attributes == ['all']:
        attributes = list(spec.db_block_keys)
    else:
        attributes = [
            attribute for token in attributes for attribute in token.split(',')
        ]

    # determine blocks
    # export_blocks = await cli_utils.async_resolve_block_range(blocks)
    export_blocks = await cli_utils.async_parse_block_slice(blocks, n=n)

    # print summary
    styles = cli.get_cli_styles()
    toolstr.print_text_box('Blocks Data', style=styles['title'])
    cli.print_bullet(key='n_blocks', value=len(export_blocks))
    cli.print_bullet(key='min block', value=min(export_blocks))
    cli.print_bullet(key='max block', value=max(export_blocks))

    blocks_data = await rpc.async_batch_eth_get_block_by_number(
        block_numbers=export_blocks,
        include_full_transactions=True,
        context=dict(provider=provider),
    )

    if export == 'stdout':
        cli.print_bullet(key='attributes', value='')
        for attribute in attributes:
            cli.print_bullet(value=attribute, indent=4)
        print()

        aliases = {
            'number': 'block',
            'median_gas': 'median gas',
        }
        rows = []
        for block in blocks_data:
            row: list[typing.Any] = []
            for attribute in attributes:
                if attribute == 'timestamp':
                    row.append(
                        tooltime.timestamp_to_iso_pretty(block['timestamp'])
                    )
                elif attribute == 'age':
                    age = tooltime.get_age(
                        block['timestamp'], 'TimelengthPhrase'
                    )
                    age = ', '.join(age.split(', ')[:2])
                    row.append(age)
                elif attribute == 'median_gas':
                    median_gas = evm.compute_transactions_gas_stats(
                        block['transactions'], normalize=True
                    )
                    row.append(median_gas)
                else:
                    row.append(block[attribute])
            rows.append(row)
        labels = [aliases.get(attribute, attribute) for attribute in attributes]
        print()
        toolstr.print_table(
            rows,
            labels=labels,
            border=styles['comment'],
            label_style=styles['title'],
            column_styles={
                'block': styles['description'],
            },
            column_formats={
                'block': {'commas': False},
                'median gas': {'postfix': ' gwei', 'trailing_zeros': True},
            },
            compact=4,
        )
    else:

        # format as dataframe
        df = pl.DataFrame(blocks_data)

        # # special attribute: time
        # if 'time' in attributes:
        #     df['time'] = df['timestamp'].map(tooltime.timestamp_to_iso)

        # # trim attributes
        # if len(attributes) > 0:
        #     df = df[attributes]

        # # output data
        # if 'number' in df:
        #     df = df.set_index('number')
        cli_utils.output_data(df, output=export, overwrite=overwrite)




from __future__ import annotations

from ctc import spec

from .. import rpc_constructors
from .. import rpc_digestors
from .. import rpc_request


def sync_web3_client_version(
    *,
    context: spec.Context = None,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_web3_client_version()
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_web3_client_version(response=response)


def sync_web3_sha3(
    data: spec.BinaryData,
    *,
    context: spec.Context = None,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_web3_sha3(data)
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_web3_sha3(response=response)


def sync_net_version(
    *,
    context: spec.Context = None,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_net_version()
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_net_version(response=response)


def sync_net_peer_count(
    *,
    context: spec.Context = None,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_net_peer_count()
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_net_peer_count(response=response)


def sync_net_listening(
    *,
    context: spec.Context = None,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_net_listening()
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_net_listening(response=response)


def sync_eth_protocol_version(
    *,
    context: spec.Context = None,
    decode_response: bool = True,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_protocol_version()
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_protocol_version(
        response=response,
        decode_response=decode_response,
    )


def sync_eth_syncing(
    *,
    context: spec.Context = None,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_syncing()
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_syncing(
        response=response,
        snake_case_response=snake_case_response,
    )


def sync_eth_chain_id(
    *,
    context: spec.Context = None,
    decode_response: bool = True,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_chain_id()
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_chain_id(
        response=response,
        decode_response=decode_response,
    )




from __future__ import annotations

import typing

from ctc import spec
from . import consensus_spec


def _get_consensus_node_host(context: spec.Context) -> str:
    from ctc import config

    chain_id = config.get_context_chain_id(context)
    if chain_id == 1:
        return 'http://testing.mainnet.beacon-api.nimbus.team'
    elif chain_id == 5:
        return 'http://unstable.prater.beacon-api.nimbus.team'
    else:
        raise Exception(
            'no known consensus node for chain_id = ' + str(chain_id)
        )


def _build_consensus_url(
    endpoint: str,
    *,
    context: spec.Context,
    params: typing.Mapping[str, typing.Any] | None = None,
) -> str:
    host = _get_consensus_node_host(context=context)
    if params is None:
        params = {}
    return host + consensus_spec.consensus_endpoints[endpoint].format(**params)


async def async_beacon_request(
    endpoint: str,
    *,
    params: typing.Mapping[str, typing.Any] | None = None,
    context: spec.Context = None,
) -> typing.Any:
    import aiohttp

    url = _build_consensus_url(
        endpoint=endpoint, context=context, params=params
    )
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            data = await response.json()
            return data['data']


def _sync_beacon_request(
    endpoint: str,
    *,
    params: typing.Mapping[str, typing.Any] | None = None,
    context: spec.Context = None,
) -> typing.Any:
    import requests

    url = _build_consensus_url(
        endpoint=endpoint, context=context, params=params
    )
    response = requests.get(url)
    return response.json()['data']




import asyncio
import os
import tempfile

import pytest

from ctc import rpc


@pytest.fixture(scope="session")
def event_loop(request):
    """create an instance of the default event loop for each test case

    adapted from https://stackoverflow.com/a/66225169
    """
    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope='session', autouse=True)
async def setup_teardown():

    # setup
    pass

    # transition to teardown
    yield

    # teardown
    await rpc.async_close_http_session()


def get_test_db_config():
    tempdir = tempfile.mkdtemp()
    return {
        'dbms': 'sqlite',
        'path': os.path.join(tempdir, 'example.db'),
    }




from __future__ import annotations

import typing

from ctc import evm
from ctc import spec
from . import block_crud


async def async_block_number_to_int(
    block: spec.BlockNumberReference,
    *,
    context: spec.Context = None,
) -> int:
    """resolve block number reference to int (e.g. converting 'latest' to int)

    Examples: 'latest', or 9999.0, or 9999
    """
    if block is None:
        block = 'latest'
    if block in spec.block_number_names:
        return await block_crud.async_get_latest_block_number(context=context)
    else:
        return evm.raw_block_number_to_int(block)


async def async_block_numbers_to_int(
    blocks: typing.Sequence[spec.BlockNumberReference],
    *,
    context: spec.Context = None,
) -> typing.Sequence[int]:
    """convert block numers to integer"""

    import asyncio

    coroutines = [
        async_block_number_to_int(block=block, context=context)
        for block in blocks
    ]
    return await asyncio.gather(*coroutines)


async def async_block_reference_to_int(
    block: spec.BlockReference,
    *,
    context: spec.Context,
) -> int:
    """convert block number, block hash, or block name to int"""
    if spec.is_block_number_reference(block):
        return await async_block_number_to_int(block, context=context)
    elif spec.is_block_hash(block):
        import ctc.rpc

        block_data: spec.RPCBlock = await ctc.rpc.async_eth_get_block_by_hash(
            block, context=context
        )
        return block_data['number']
    else:
        raise Exception('unknown block format: ' + str(block))


async def async_block_references_to_int(
    blocks: typing.Sequence[spec.BlockReference],
    *,
    context: spec.Context,
) -> typing.Sequence[int]:
    """convert block number, block hash, or block name to int"""
    import asyncio

    coroutines = [
        async_block_reference_to_int(block=block, context=context)
        for block in blocks
    ]
    return await asyncio.gather(*coroutines)




from __future__ import annotations

import os
import typing

if typing.TYPE_CHECKING:

    from typing_extensions import TypedDict

    DirectoryTree = typing.Mapping[str, typing.Mapping[str, typing.Any]]

    class DataDirSpec(TypedDict, total=False):
        files: typing.Sequence[str]
        directory_tree: DirectoryTree

        # # for now, only specifies entries in root data dir
        # files: list[str]  # list of expected file paths
        # directories: list[str]  # list of expected directory paths
        # directory_subdirs: typing.Mapping[
        #     str, typing.Sequence[str]
        # ]  # list of items in each directory
        # move_items: typing.Mapping[str, str]

    DataSpecVersion = str


data_spec_order = [
    '0.2.0',
    '0.3.0',
    '0.3.1',
    '0.3.2',
    '0.3.3',
    '0.3.4',
    '0.3.5',
    '0.3.6',
]

_0_3_0_spec: DataDirSpec = {
    'directory_tree': {
        'dbs': {},
        'logs': {'rpc': {}, 'db': {}},
        'evm': {},
    },
    'files': [
        './directory_version',
    ],
}

data_dir_specs: typing.Mapping[DataSpecVersion, DataDirSpec] = {
    '0.2.0': {
        'files': [],
        'directory_tree': {},
    },
    '0.3.0': _0_3_0_spec,
    '0.3.1': _0_3_0_spec,
    '0.3.2': _0_3_0_spec,
    '0.3.3': _0_3_0_spec,
    '0.3.4': _0_3_0_spec,
    '0.3.5': _0_3_0_spec,
    '0.3.6': _0_3_0_spec,
}


def initialize_data_subdirs(data_dir: str, *, version: str) -> None:
    if version in data_spec_order:
        data_dir_spec = data_dir_specs[version]
    else:
        raise Exception('invalid version: ' + str(version))

    directory_tree = data_dir_spec['directory_tree']
    tree_paths = _flatten_tree(directory_tree)
    for tree_path in tree_paths:
        os.makedirs(os.path.join(data_dir, tree_path), exist_ok=True)


def _flatten_tree(directory_tree: DirectoryTree) -> typing.Sequence[str]:
    """return depth-first flattening of nested dict filesystem tree"""
    flat = []
    for dirname, subdirs in directory_tree.items():
        flat.append(dirname)
        if len(subdirs) > 0:
            for subdir in _flatten_tree(subdirs):
                flat.append(os.path.join(dirname, subdir))
    return flat


# def initialize_data_subdirs(data_dir: str, *, version: DataSpecVersion) -> None:
#     data_dir_spec = data_dir_specs[version]

#     for dirname in data_dir_spec.get('directories', []):
#         path = os.path.join(data_dir, dirname)
#         if not os.path.isdir(path):
#             print('creating directory:', path)
#             os.makedirs(path, exist_ok=True)

#     for dirname in data_dir_spec.get('directory_subdirs', []):
#         for subdirname in data_dir_spec['directory_subdirs'][dirname]:
#             path = os.path.join(data_dir, dirname, subdirname)
#             if not os.path.isdir(path):
#                 print('creating directory:', path)
#                 os.makedirs(path, exist_ok=True)


# def get_data_dir_version(data_dir: str | None = None) -> DataSpecVersion:
#     if data_dir is None:
#         data_dir = config.get_data_dir()

#     version_file = os.path.join(data_dir, 'directory_version')

#     if not os.path.isfile(version_file):
#         return '0.2.0'
#     else:
#         with open(version_file, 'r') as f:
#             version = f.read()
#         if version in data_spec_order:
#             return typing.cast(DataSpecVersion, version)
#         else:
#             raise Exception('unknown data_spec_version: ' + str(version))


# def fully_migrate_data_dir(data_dir: str | None = None) -> None:

#     if data_dir is None:
#         data_dir = config.get_data_dir()

#     # detect current version
#     current_version = get_data_dir_version()
#     latest_version = data_spec_order[-1]
#     if current_version == latest_version:
#         print('data_dir already fully migrated')
#         return

#     # gather all migrate functions, use None if no migration necessary
#     migrate_functions = {
#         '0.3.0': migrate_data_dir__0_2_0__to__0_3_0,
#         '0.3.1': None,
#     }

#     # perform each upgrade function sequentially
#     index = data_spec_order.index(current_version)
#     steps = data_spec_order[index + 1:]
#     for step in steps:
#         f = migrate_functions[step]
#         if f is not None:
#             f(data_dir=data_dir)


# def migrate_data_dir__0_2_0__to__0_3_0(
#     data_dir: str,
#     *,
#     delete_old_data: bool = True,
#     confirm_delete: bool = False,
# ) -> None:

#     data_dir_spec = data_dir_specs['0.3.0']

#     # create version file
#     version_file = os.path.join(data_dir, 'directory_version')
#     print('creating new version indicator file:', version_file)
#     with open(version_file, 'w') as f:
#         f.write('0.3.0')

#     # create new directories
#     for relpath in data_dir_spec['directories']:
#         dirpath = os.path.join(data_dir, relpath)
#         print('creating directory:', dirpath)
#         os.makedirs(dirpath, exist_ok=True)

#     # migrate data (the only old data that is migrated is events)
#     print()
#     for old_item, new_item in data_dir_spec['move_items'].items():
#         old_path = os.path.join(data_dir, old_item)
#         new_path = os.path.join(data_dir, new_item)
#         if os.path.exists(old_path):
#             print('moving', old_item, 'to', new_item)
#             new_parent = os.path.dirname(new_path)
#             os.makedirs(new_parent, exist_ok=True)
#             shutil.move(old_path, new_path)

#     # delete old files
#     if delete_old_data:
#         to_delete = []
#         for item in os.listdir(data_dir):
#             if (
#                 item not in data_dir_spec['directories']
#                 and item not in data_dir_spec['files']
#             ):
#                 to_delete.append(item)
#             elif (
#                 item in data_dir_spec['directories']
#                 and item in data_dir_spec['directory_subdirs']
#             ):
#                 directory_subdirs = data_dir_spec['directory_subdirs'][item]
#                 for subitem in os.listdir(os.path.join(data_dir, item)):
#                     if subitem not in directory_subdirs:
#                         to_delete.append(os.path.join(item, subitem))
#         if len(to_delete) > 0:
#             if not confirm_delete:
#                 print()
#                 print(
#                     'the following files and directories are not used in ctc 0.3.0:'
#                 )
#                 for item in sorted(to_delete):
#                     print('-', item)
#                 print()
#                 if not toolcli.input_yes_or_no('delete these items? '):
#                     raise Exception(
#                         'migration unfinished, must delete old files'
#                     )

#             for path in to_delete:
#                 path = os.path.join(data_dir, path)
#                 if os.path.isfile(path) or os.path.islink(path):
#                     os.remove(path)
#                 elif os.path.isdir(path):
#                     shutil.rmtree(path)
#                 else:
#                     raise Exception('cannot process path: ' + str(path))

#     print('data migration complete, now using the 0.3.0 data schema')



from __future__ import annotations

import typing

from ctc import spec

from .. import dex_class
from .. import dex_class_utils

if typing.TYPE_CHECKING:
    import tooltime


async def async_get_pools(
    *,
    dex: typing.Type[dex_class.DEX] | str | None = None,
    factory: spec.Address | None = None,
    assets: typing.Sequence[spec.Address] | None = None,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    update: bool = False,
    context: spec.Context = None,
) -> typing.Sequence[spec.DexPool]:
    """get DEX pools matching given inputs"""

    # get dex
    all_dex_factories = dex is not None and factory is None
    if factory is None and dex is None:
        dex = dex_class.DEX

        if update:
            await async_update_all_dexes(context=context)
            update = False

    else:
        dex = dex_class_utils.get_dex_class(
            dex=dex,
            factory=factory,
            context=context,
        )

    return await dex.async_get_pools(
        factory=factory,
        all_dex_factories=all_dex_factories,
        assets=assets,
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        update=update,
        context=context,
    )


async def async_update_all_dexes(
    context: spec.Context = None,
) -> typing.Mapping[str, typing.Mapping[str, typing.Any]]:
    """update local DEX database with latest on-chain entries"""

    import asyncio

    all_dexes = dex_class_utils.get_all_dex_classes()

    coroutines = []
    for dex in all_dexes.values():
        coroutine = dex.async_update_pools(context=context)
        coroutines.append(coroutine)

    results = await asyncio.gather(*coroutines)

    updates = {}
    for dex_name, result in zip(all_dexes.keys(), results):
        updates[dex_name] = {'new_pools': result}

    return updates



from __future__ import annotations

import os
import typing

from .. import transaction_utils
from . import block_times
from . import block_normalize
from ctc import evm
from ctc import spec

if typing.TYPE_CHECKING or os.environ.get('BUILDING_SPHINX') == '1':

    from typing_extensions import TypedDict
    from ctc import db

    class BlockGasStats(TypedDict):
        min_gas_price: int | float | None
        median_gas_price: int | float | None
        mean_gas_price: float | None
        max_gas_price: int | float | None
        n_transactions: int
        # base_fee: int | float | None
        # gas_used: int
        # gas_limit: int


#
# # query median gas fees
#


async def async_get_median_block_gas_fee(
    block: spec.BlockNumberReference,
    *,
    normalize: bool = True,
    context: spec.Context = None,
) -> db.BlockGasRow:
    """get median gas fee for a given block"""

    from ctc import config
    from ctc import rpc

    block = await evm.async_block_number_to_int(block, context=context)

    read_cache, write_cache = config.get_context_cache_read_write(
        schema_name='block_gas', context=context
    )
    if read_cache:
        from ctc import db

        try:
            result = await db.async_query_median_block_gas_fee(
                block, context=context
            )

            if result is not None:
                if normalize and result['median_gas_fee'] is not None:
                    result['median_gas_fee'] /= 1e9

                return result

        except Exception:
            pass

    block_timestamp = await block_times.async_get_block_timestamp(
        block=block, context=context
    )
    txs = await transaction_utils.async_get_block_transactions(
        block, context=context
    )
    return {
        'block_number': block,
        'timestamp': block_timestamp,
        'median_gas_fee': compute_transactions_median_gas_fee(
            txs,
            normalize=normalize,
        ),
    }


async def async_get_median_blocks_gas_fees(
    blocks: typing.Sequence[spec.BlockNumberReference],
    *,
    normalize: bool = True,
    verbose: bool = True,
    latest_block_number: int | None = None,
    context: spec.Context = None,
) -> typing.Sequence[db.BlockGasRow]:
    """get median gas fees for multiple blocks"""

    from ctc import config

    blocks = await evm.async_block_numbers_to_int(blocks)

    # get data from db
    read_cache, write_cache = config.get_context_cache_read_write(
        schema_name='block_gas', context=context
    )
    if read_cache:
        from ctc import db

        result = await db.async_query_median_blocks_gas_fees(
            blocks, context=context
        )

        if result is None:
            fee_map: typing.MutableMapping[int, db.BlockGasRow] = {}
            missing: typing.Sequence[int] = blocks
        else:
            fee_map = dict(result)
            missing = []
            for block in blocks:
                if block not in fee_map:
                    missing.append(block)
    else:
        fee_map = {}

    # get data from rpc
    if len(missing) > 0:
        if verbose:
            import toolstr
            from ctc import cli

            styles = cli.get_cli_styles()
            toolstr.print(
                'fetching tx gas data for '
                + toolstr.add_style(
                    str(len(missing)), styles['description'] + ' bold'
                )
                + ' blocks',
            )
            print()

        from ctc import config

        context = config.update_context(
            context, merge_provider={'chunk_size': 1}
        )

        # get txs by block
        txs = await transaction_utils.async_get_blocks_transactions(
            blocks=missing,
            context=context,
        )
        blocks_timestamps = await block_times.async_get_block_timestamps(
            blocks=blocks, context=context
        )
        txs_by_block: dict[int, list[spec.DBTransaction]] = {}
        for tx in txs:
            block = tx['block_number']
            txs_by_block.setdefault(block, [])
            txs_by_block[block].append(tx)

        for block, timestamp in zip(missing, blocks_timestamps):
            fee = compute_transactions_median_gas_fee(
                txs_by_block[block], normalize=normalize
            )
            fee_map[block] = {
                'block_number': block,
                'median_gas_fee': fee,
                'timestamp': timestamp,
            }

    if normalize:
        for block, fee_data in fee_map.items():
            if fee_data is not None and fee_data['median_gas_fee'] is not None:
                fee_data['median_gas_fee'] /= 1e9

    return [fee_map[block] for block in blocks]


#
# # query gas stats
#


async def async_get_block_gas_stats(
    block: spec.BlockNumberReference,
    *,
    normalize: bool = True,
    context: spec.Context = None,
) -> BlockGasStats:
    """get gas usage statistics for a given block"""

    block_number = await block_normalize.async_block_reference_to_int(
        block=block, context=context
    )
    txs = await transaction_utils.async_get_block_transactions(
        block=block_number, context=context
    )
    return compute_transactions_gas_stats(txs, normalize=normalize)


async def async_get_gas_stats_by_block(
    blocks: typing.Sequence[spec.BlockNumberReference],
    *,
    normalize: bool = True,
    context: spec.Context = None,
) -> list[BlockGasStats]:
    """get block gas usage statistics of multiple blocks"""

    # get txs of blocks
    int_blocks = await block_normalize.async_block_numbers_to_int(
        blocks=blocks, context=context
    )
    txs = await transaction_utils.async_get_blocks_transactions(
        blocks=int_blocks, context=context
    )

    # sort txs by block
    txs_by_block: dict[int, list[spec.DBTransaction]] = {}
    for tx in txs:
        block = tx['block_number']
        txs_by_block.setdefault(block, [])
        txs_by_block[block].append(tx)

    # compute stats per block
    return [
        compute_transactions_gas_stats(
            txs_by_block[int_block], normalize=normalize
        )
        for int_block in int_blocks
    ]


#
# # compute gas stats
#


def compute_transactions_median_gas_fee(
    transactions: typing.Sequence[spec.DBTransaction],
    *,
    normalize: bool,
) -> int | float | None:
    """compute median gas fee of transactions in block"""

    import numpy as np

    # gather gas fees
    gas_fees = []
    for transaction in transactions:
        if isinstance(transaction, str):
            raise Exception(
                'must use a block with include_full_transactions=True'
            )
        gas_fees.append(transaction['gas_price'])

    # compute median
    median = float(np.median(gas_fees))

    # # normalize
    # if normalize:
    #     median = median / 1e9

    return median


def compute_transactions_gas_stats(
    transactions: typing.Sequence[spec.DBTransaction],
    *,
    normalize: bool = True,
) -> BlockGasStats:
    """compute gas usage statistics for given block"""

    import numpy as np

    if len(transactions) > 0:

        gas_prices: list[int | float] = [
            transaction['gas_price'] for transaction in transactions
        ]

        if normalize:
            gas_prices = [gas_price / 1e9 for gas_price in gas_prices]

        min_gas_price = min(gas_prices)
        median_gas_price = float(np.median(gas_prices))
        mean_gas_price = sum(gas_prices) / len(gas_prices)
        max_gas_price = max(gas_prices)

    else:
        min_gas_price = None
        median_gas_price = None
        mean_gas_price = None
        max_gas_price = None

    return {
        'min_gas_price': min_gas_price,
        'median_gas_price': median_gas_price,
        'mean_gas_price': mean_gas_price,
        'max_gas_price': max_gas_price,
        'n_transactions': len(transactions),
    }


# def compute_block_gas_stats(block: spec.DBBlock, normalize: bool = False):
#     base_fee: int | float | None = block.get('base_fee_per_gas')
#     if normalize:
#         if base_fee is not None:
#             base_fee = base_fee / 1e9
#     return {
#         'base_fee': base_fee,
#         'gas_used': block['gas_used'],
#         'gas_limit': block['gas_limit'],
#     }




from __future__ import annotations

import typing

from ctc import evm
from ctc import spec

if typing.TYPE_CHECKING:
    import tooltime


@typing.overload
async def async_resolve_block_range(
    *,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    allow_none: typing.Literal[False],
    context: spec.Context = None,
    to_int: typing.Literal[True],
    start_none_means: spec.BlockNumberReference | None = None,
    end_none_means: spec.BlockNumberReference | None = None,
) -> tuple[int, int]:
    ...


@typing.overload
async def async_resolve_block_range(
    *,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    allow_none: typing.Literal[False],
    context: spec.Context = None,
    to_int: typing.Literal[False] = False,
    start_none_means: spec.BlockNumberReference | None = None,
    end_none_means: spec.BlockNumberReference | None = None,
) -> tuple[spec.BlockNumberReference, spec.BlockNumberReference]:
    ...


@typing.overload
async def async_resolve_block_range(
    *,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    allow_none: bool,
    context: spec.Context = None,
    to_int: typing.Literal[True],
    start_none_means: spec.BlockNumberReference | None = None,
    end_none_means: spec.BlockNumberReference | None = None,
) -> tuple[int | None, int | None]:
    ...


@typing.overload
async def async_resolve_block_range(
    *,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    allow_none: bool,
    context: spec.Context = None,
    to_int: bool = False,
    start_none_means: spec.BlockNumberReference | None = None,
    end_none_means: spec.BlockNumberReference | None = None,
) -> tuple[spec.BlockNumberReference | None, spec.BlockNumberReference | None]:
    ...


async def async_resolve_block_range(
    *,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    allow_none: bool,
    context: spec.Context = None,
    to_int: bool = False,
    start_none_means: spec.BlockNumberReference | None = None,
    end_none_means: spec.BlockNumberReference | None = None,
) -> tuple[spec.BlockNumberReference | None, spec.BlockNumberReference | None]:
    """resolve block or timestamp range to a block range"""

    import asyncio

    if start_block is None and start_time is None:
        start_block = start_none_means
    if end_block is None and end_time is None:
        end_block = end_none_means

    tasks = {}
    if start_block is not None:
        if to_int:
            tasks['start_block'] = asyncio.create_task(
                evm.async_block_number_to_int(start_block, context=context)
            )
        else:
            start_block = start_block
    elif start_time is not None:
        tasks['start_block'] = asyncio.create_task(
            evm.async_get_block_of_timestamp(start_time, context=context)
        )
    else:
        if allow_none:
            start_block = None
        else:
            raise Exception('must specify start_block or start_time')

    if end_block is not None:
        if to_int:
            tasks['end_block'] = asyncio.create_task(
                evm.async_block_number_to_int(end_block, context=context)
            )
        else:
            end_block = end_block
    elif end_time is not None:
        tasks['end_block'] = asyncio.create_task(
            evm.async_get_block_of_timestamp(end_time, context=context)
        )
    else:
        if allow_none:
            end_block = None
        else:
            raise Exception('must specify end_block or end_time')

    if 'start_block' in tasks:
        start_block = await tasks['start_block']
    if 'end_block' in tasks:
        end_block = await tasks['end_block']

    return start_block, end_block



from .abi_utils import *
from .address_utils import *
from .binary_utils import *
from .block_utils import *
from .contract_utils import *
from .erc20_utils import *
from .erc721_utils import *
from .erc4626_utils import *
from .eth_utils import *
from .event_utils import *
from .network_utils import *
from .transaction_utils import *
from .trace_utils import *




from __future__ import annotations

import typing

from ctc import evm
from ctc import rpc
from ctc import spec
from . import curve_spec

if typing.TYPE_CHECKING:
    import tooltime


async def async_get_pool_A(
    pool: spec.Address,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
) -> int:

    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=curve_spec.pool_function_abis['A'],
        block_number=block,
        context=context,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def async_get_pool_future_A_time(
    pool: spec.Address,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
) -> int:
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=curve_spec.pool_function_abis['future_A_time'],
        block_number=block,
        context=context,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def async_get_pool_initial_A(
    pool: spec.Address,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
) -> int:
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=curve_spec.pool_function_abis['initial_A'],
        block_number=block,
        context=context,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def async_get_pool_initial_A_time(
    pool: spec.Address,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
) -> int:
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=curve_spec.pool_function_abis['initial_A_time'],
        block_number=block,
        context=context,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def async_get_pool_future_A(
    pool: spec.Address,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
) -> int:
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=curve_spec.pool_function_abis['future_A'],
        block_number=block,
        context=context,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def async_get_pool_ramps(
    *,
    context: spec.Context = None,
) -> spec.DataFrame:
    """get Ramp events"""
    raise NotImplementedError()


async def async_get_A_history(
    pool: spec.Address,
    *,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    context: spec.Context = None,
) -> typing.Sequence[float]:
    """get history of pool's A parameter"""

    import asyncio
    import numpy as np
    from ctc.toolbox import pl_utils

    start_block, end_block = await evm.async_resolve_block_range(
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        allow_none=True,
        context=context,
    )

    # get ramp events
    pool_start_block = await evm.async_get_contract_creation_block(
        pool,
        verbose=False,
        context=context,
    )
    latest_block = await evm.async_get_latest_block_number(context=context)
    blocks: spec.NumpyArray = np.arange(start_block, end_block, dtype=int)
    block_timestamps_task = asyncio.create_task(
        evm.async_get_block_timestamps(
            blocks=typing.cast(typing.Sequence[int], blocks),
            context=context,
        )
    )

    event_abi: spec.EventABI = {
        'anonymous': False,
        'inputs': [
            {
                'indexed': False,
                'name': 'old_A',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'name': 'new_A',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'name': 'initial_time',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'name': 'future_time',
                'type': 'uint256',
            },
        ],
        'name': 'RampA',
        'type': 'event',
    }

    events = await evm.async_get_events(
        contract_address=pool,
        event_abi=event_abi,
        start_block=pool_start_block,
        end_block=latest_block,
        context=context,
    )
    events = events[
        ['arg__old_A', 'arg__new_A', 'arg__initial_time', 'arg__future_time']
    ]

    # initial deployment parameters
    initial_A = await async_get_pool_initial_A(
        pool, block=pool_start_block, context=context
    )
    initial_A_time = 0
    events.loc[start_block, 0, 0] = [initial_A, initial_A, initial_A_time, 0]  # type: ignore
    events = events.sort('block_number', 'log_index')

    # interpolate per block
    events = pl_utils.interpolate(
        events,
        index_column='block_number',
        end_index=latest_block,
    )

    # get blocks timestamps
    block_timestamps = await block_timestamps_task

    # need per-block timestamps over a long time range to continue :-\
    A = compute_A(
        initial_A=typing.cast(
            typing.Sequence[typing.Union[int, float]], events['arg__old_A']
        ),
        future_A=typing.cast(
            typing.Sequence[typing.Union[int, float]], events['arg__new_A']
        ),
        initial_A_time=typing.cast(
            typing.Sequence[typing.Union[int, float]],
            events['arg__initial_time'],
        ),
        future_A_time=typing.cast(
            typing.Sequence[typing.Union[int, float]],
            events['arg__future_time'],
        ),
        timestamps=block_timestamps,
    )

    return A


def compute_A(
    *,
    initial_A: typing.Sequence[int | float],
    initial_A_time: typing.Sequence[int | float],
    future_A: typing.Sequence[int | float],
    future_A_time: typing.Sequence[int | float],
    timestamps: typing.Sequence[int | float],
) -> typing.Sequence[float]:
    """perform linear interpolation between initial and future A values

    TODO: need to convert to int?
    """

    import numpy as np

    result = _compute_A(
        initial_A=np.array(initial_A),
        initial_A_time=np.array(initial_A_time),
        future_A=np.array(future_A),
        future_A_time=np.array(future_A_time),
        timestamps=np.array(timestamps),
    )

    return list(result)


def _compute_A(
    *,
    initial_A: spec.NumpyArray,
    initial_A_time: spec.NumpyArray,
    future_A: spec.NumpyArray,
    future_A_time: spec.NumpyArray,
    timestamps: spec.NumpyArray,
) -> spec.NumpyArray:

    intercept = initial_A
    slope = (future_A - initial_A) / (future_A_time - initial_A_time)
    result = intercept + slope * (timestamps - initial_A_time)

    mask = timestamps > future_A_time
    result[mask] = future_A

    if typing.TYPE_CHECKING:
        return typing.cast(spec.NumpyArray, result)
    else:
        return result




from __future__ import annotations

import typing

from ctc import spec

from .. import rpc_constructors
from .. import rpc_digestors
from .. import rpc_request


async def async_eth_block_number(
    *,
    context: spec.Context = None,
    decode_response: bool = True,
    raw_output: bool = False,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_block_number()
    response = await rpc_request.async_send(
        request,
        context=context,
        raw_output=raw_output,
    )

    if raw_output:
        return response
    else:
        return rpc_digestors.digest_eth_block_number(
            response=response,
            decode_response=decode_response,
        )


async def async_eth_get_block_by_hash(
    block_hash: str,
    *,
    include_full_transactions: bool = False,
    context: spec.Context = None,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_get_block_by_hash(
        block_hash=block_hash,
        include_full_transactions=include_full_transactions,
    )
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_eth_get_block_by_hash(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
    )


async def async_eth_get_block_by_number(
    block_number: spec.StandardBlockNumber,
    *,
    include_full_transactions: bool = False,
    decode_response: bool = True,
    context: spec.Context = None,
    snake_case_response: bool = True,
    raw_output: bool = False,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_get_block_by_number(
        block_number=block_number,
        include_full_transactions=include_full_transactions,
    )
    response = await rpc_request.async_send(
        request,
        context=context,
        raw_output=raw_output,
    )

    if raw_output:
        return response
    else:
        return rpc_digestors.digest_eth_get_block_by_number(
            response=response,
            decode_response=decode_response,
            snake_case_response=snake_case_response,
        )


async def async_eth_get_uncle_count_by_block_hash(
    block_hash: str,
    *,
    context: spec.Context = None,
    decode_response: bool = True,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_get_uncle_count_by_block_hash(
        block_hash=block_hash,
    )
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_eth_get_uncle_count_by_block_hash(
        response=response,
        decode_response=decode_response,
    )


async def async_eth_get_uncle_count_by_block_number(
    block_number: spec.StandardBlockNumber,
    *,
    context: spec.Context = None,
    decode_response: bool = True,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_get_uncle_count_by_block_number(
        block_number=block_number,
    )
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_eth_get_uncle_count_by_block_number(
        response=response,
        decode_response=decode_response,
    )


async def async_eth_get_uncle_by_block_hash_and_index(
    block_hash: str,
    uncle_index: str,
    *,
    context: spec.Context = None,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_get_uncle_by_block_hash_and_index(
        block_hash=block_hash,
        uncle_index=uncle_index,
    )
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_eth_get_uncle_by_block_hash_and_index(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
    )


async def async_eth_get_uncle_by_block_number_and_index(
    block_number: spec.StandardBlockNumber,
    uncle_index: str,
    *,
    context: spec.Context = None,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    request = (
        rpc_constructors.construct_eth_get_uncle_by_block_number_and_index(
            block_number=block_number,
            uncle_index=uncle_index,
        )
    )
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_eth_get_uncle_by_block_number_and_index(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
    )


async def async_eth_fee_history(
    block_number: spec.BlockNumberReference = 'latest',
    block_count: int = 1024,
    *,
    reward_percentiles: typing.Sequence[float] | None = None,
    context: spec.Context = None,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    request = (
        rpc_constructors.construct_eth_fee_history(
            block_number=block_number,
            block_count=block_count,
        )
    )
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_eth_fee_history(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
    )




from __future__ import annotations

import asyncio
import typing

import toolstr

from ctc import cli
from ctc import evm
from ctc import spec
from . import safe_events
from . import safe_metadata


async def async_print_safe_summary(
    address: spec.Address,
    *,
    verbose: bool = False,
    context: spec.Context = None,
) -> None:
    import tooltime

    owners_coroutine = safe_metadata.async_get_safe_owners(
        address, context=context
    )
    threshold_coroutine = safe_metadata.async_get_safe_threshold(
        address, context=context
    )
    nonce_coroutine = safe_metadata.async_get_safe_nonce(
        address, context=context
    )
    creation_block_coroutine = evm.async_get_contract_creation_block(
        address, context=context
    )

    owners, threshold, nonce, creation_block = await asyncio.gather(
        owners_coroutine,
        threshold_coroutine,
        nonce_coroutine,
        creation_block_coroutine,
    )

    if creation_block is not None:
        creation_timestamp = await evm.async_get_block_timestamp(
            creation_block, context=context
        )
        creation_date = tooltime.timestamp_to_iso_pretty(creation_timestamp)
        age = tooltime.get_age(creation_timestamp, 'TimelengthPhrase')
    else:
        creation_timestamp = None
        creation_date = None
        age = None

    styles = cli.get_cli_styles()

    toolstr.print_text_box('Gnosis safe ' + str(address), style=styles['title'])
    cli.print_bullet(
        key='threshold',
        value=str(threshold)
        + toolstr.add_style(' / ', styles['comment'])
        + str(len(owners)),
    )
    cli.print_bullet(key='owners', value='')
    for owner in owners:
        cli.print_bullet(value=owner, indent=4)
    cli.print_bullet(key='nonce', value=nonce)
    cli.print_bullet(key='creation block', value=creation_block)
    cli.print_bullet(key='creation date', value=creation_date)
    cli.print_bullet(key='age', value=age)

    if verbose:
        print()
        print()
        await async_print_safe_executions(address)
        print()
        print()
        await async_print_safe_owner_history(address)
        print()
        print()
        await async_print_safe_erc20s(address)


async def async_print_safe_erc20s(
    address: spec.Address, *, context: spec.Context = None
) -> None:
    styles = cli.get_cli_styles()

    default_erc20s = await evm.async_get_default_erc20_tokens(context=context)
    erc20_addresses = [default['address'] for default in default_erc20s]

    toolstr.print_text_box('Common ERC20s in safe', style=styles['title'])
    balances = await evm.async_get_erc20s_balances(
        wallet=address,
        tokens=erc20_addresses,
        context=context,
    )

    rows = []
    for erc20, balance in zip(default_erc20s, balances):
        if balance > 0:
            row = [erc20['symbol'], balance]
            rows.append(row)
    labels = ['token', 'balance']
    if len(rows) == 0:
        print('[none]')
    else:
        print()
        toolstr.print_table(
            rows,
            labels=labels,
            indent=4,
            border=styles['comment'],
            label_style=styles['title'],
            column_styles={
                'token': styles['option'],
                'balance': styles['description'],
            },
        )


async def async_print_safe_executions(
    address: spec.Address, *, context: spec.Context = None
) -> None:
    executions = await safe_events.async_get_safe_executions(
        address, context=context
    )

    # # TODO: output more data about each transaction
    # transaction_hashes = list(executions['transaction_hash'].unique())
    # transactions = await evm.async_get_transactions(transaction_hashes)

    styles = cli.get_cli_styles()

    labels = [
        'nonce',
        'block',
        #     'timestamp',
        #     'age',
        'transaction',
        #     'safeTransaction',
    ]

    rows = []
    for i in range(len(executions)):
        row = [i, executions['block_number'][i]]
        row.append(executions['transaction_hash'][i])
        rows.append(row)

    toolstr.print_text_box('Safe Executions', style=styles['title'])
    print()
    toolstr.print_table(
        rows,
        labels=labels,
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'nonce': styles['option'],
            'block': styles['description'],
            'transaction': styles['metavar'],
        },
        compact=2,
    )


async def async_print_safe_owner_history(
    address: spec.Address,
    *,
    end_block: spec.BlockReference | None = None,
    context: spec.Context = None,
) -> None:
    if end_block is None:
        end_block = await evm.async_get_latest_block_number(context=context)

    safe_setup_coroutine = safe_events.async_get_safe_setup(
        address, context=context
    )
    owner_adds_coroutine = safe_events.async_get_safe_owner_adds(
        address, end_block=end_block, context=context
    )
    owner_removes_coroutine = safe_events.async_get_safe_owner_removes(
        address, end_block=end_block, context=context
    )
    threshold_changes_coroutine = safe_events.async_get_safe_threshold_changes(
        address, end_block=end_block, context=context
    )

    safe_setup: typing.Any = await safe_setup_coroutine
    owner_adds = await owner_adds_coroutine
    owner_removes = await owner_removes_coroutine
    threshold_changes = await threshold_changes_coroutine

    changes: typing.Sequence[typing.Any] = (
        list(owner_adds.to_dicts())
        + list(owner_removes.to_dicts())
        + list(threshold_changes.to_dicts())
    )
    changes = sorted(changes, key=lambda change: change['block_number'])  # type: ignore

    styles = cli.get_cli_styles()

    if safe_setup is not None:
        blocks = [safe_setup.name[0]]
        owners = [set(safe_setup['arg__owners'])]
        threshold = [safe_setup['arg__threshold']]
    else:
        creation_block_coroutine = evm.async_get_contract_creation_block(
            address, context=context
        )
        creation_block = await creation_block_coroutine
        owners_coroutine = safe_metadata.async_get_safe_owners(
            address, block=creation_block, context=context
        )
        threshold_coroutine = safe_metadata.async_get_safe_threshold(
            address, block=creation_block, context=context
        )
        blocks = [creation_block]
        owners = [set(await owners_coroutine)]
        threshold = [await threshold_coroutine]
    diffs = ['safe initialized']
    for change in changes:
        blocks.append(change['block_number'])

        # AddedOwner
        if (
            change['event_hash']
            == '0x9465fa0c962cc76958e6373a993326400c1c94f8be2fe3a952adfa7f60b2ea26'
        ):
            diffs.append(
                'add owner '
                + toolstr.add_style(change['arg__owner'], styles['metavar'])
            )
            new_owners = set(owners[-1])
            new_owners.add(change['arg__owner'])
            owners.append(new_owners)
            threshold.append(threshold[-1])

        # RemovedOwner
        elif (
            change['event_hash']
            == '0xf8d49fc529812e9a7c5c50e69c20f0dccc0db8fa95c98bc58cc9a4f1c1299eaf'
        ):
            diffs.append(
                'remove owner '
                + toolstr.add_style(change['arg__owner'], styles['metavar'])
            )
            new_owners = set(owners[-1])
            new_owners.remove(change['arg__owner'])
            owners.append(new_owners)
            threshold.append(threshold[-1])

        # ChangedThreshold
        elif (
            change['event_hash']
            == '0x610f7ff2b304ae8903c3de74c60c6ab1f7d6226b3f52c5161905bb5ad4039c93'
        ):
            diffs.append(
                'change signing threshold to '
                + toolstr.add_style(
                    str(change['arg__threshold']), styles['description']
                )
            )
            owners.append(owners[-1])
            threshold.append(change['arg__threshold'])

        else:
            raise Exception()

    # convert to rows for printout
    rows = []
    for d, diff in enumerate(diffs):
        row = [
            str(blocks[d]),
            diffs[d],
            str(threshold[d])
            + toolstr.add_style(' / ', styles['comment'])
            + str(len(owners[d])),
        ]
        rows.append(row)
    labels = [
        'block',
        'diff',
        'threshold',
    ]

    toolstr.print_text_box('Safe Permission History', style=styles['title'])
    print()
    toolstr.print_table(
        rows=rows,
        labels=labels,
        compact=4,
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'block': styles['description'],
            'threshold': styles['description'],
        },
    )




from __future__ import annotations

import typing

import toolsql

from ctc import evm
from ctc import spec
from ... import schema_utils
from . import events_schema_defs


if typing.TYPE_CHECKING:
    from typing_extensions import Literal


async def async_upsert_events(
    *,
    encoded_events: typing.Sequence[spec.EncodedEvent],
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> None:
    if len(encoded_events) == 0:
        return

    # fill missing rows
    if isinstance(encoded_events[0], dict):
        for event in encoded_events:
            event.setdefault('topic1', None)
            event.setdefault('topic2', None)
            event.setdefault('topic3', None)
            event.setdefault('unindexed', None)

            if (
                event.get('block_number') is None
                or event.get('transaction_index') is None
                or event.get('log_index') is None
            ):
                raise Exception('must specify full index for event')

        as_binary = [
            evm.binarize_fields(
                encoded_event,
                events_schema_defs._event_binary_fields,
            )
            for encoded_event in encoded_events
        ]

    elif isinstance(encoded_events[0], tuple):
        as_binary = [
            event[:3]
            + tuple(
                (bytes.fromhex(value[2:]) if value is not None else None)
                for value in event[3:]
            )
            + (None,)
            for event in encoded_events
        ]

    else:
        raise Exception('invalid format for encoded_events')

    table = schema_utils.get_table_schema('events', context=context)
    await toolsql.async_insert(
        conn=conn,
        table=table,
        rows=as_binary,
        upsert=True,
    )


async def async_upsert_event_query(
    *,
    event_query: spec.EventQuery,
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> None:
    # convert to binary as needed
    as_binary = evm.binarize_fields(
        event_query,
        events_schema_defs._event_query_binary_fields,
    )

    table = schema_utils.get_table_schema('event_queries', context=context)
    await toolsql.async_insert(
        conn=conn,
        table=table,
        row=as_binary,
        upsert=True,
    )


async def async_upsert_event_queries(
    *,
    event_queries: typing.Sequence[spec.EventQuery],
    conn: toolsql.AsyncConnection,
    context: spec.Context,
) -> None:
    as_binary = [
        evm.binarize_fields(
            event_query,
            events_schema_defs._event_query_binary_fields,
        )
        for event_query in event_queries
    ]

    table = schema_utils.get_table_schema('event_queries', context=context)
    await toolsql.async_insert(
        conn=conn,
        table=table,
        rows=as_binary,
        upsert=True,
    )


@typing.overload
async def async_select_events(
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
    contract_address: spec.Address | None = None,
    event_hash: typing.Any | None = None,
    topic1: typing.Any | None = None,
    topic2: typing.Any | None = None,
    topic3: typing.Any | None = None,
    start_block: int | None = None,
    end_block: int | None = None,
    columns: typing.Sequence[str] | None = None,
    binary_output_format: Literal['binary', 'prefix_hex'] = 'prefix_hex',
    topic_output_format: Literal['binary', 'prefix_hex'] = 'prefix_hex',
) -> spec.DataFrame:
    ...


@typing.overload
async def async_select_events(
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
    contract_address: spec.Address | None = None,
    event_hash: typing.Any | None = None,
    topic1: typing.Any | None = None,
    topic2: typing.Any | None = None,
    topic3: typing.Any | None = None,
    start_block: int | None = None,
    end_block: int | None = None,
    columns: typing.Sequence[str] | None = None,
    output_format: Literal['polars'] = 'polars',
    binary_output_format: Literal['binary', 'prefix_hex'] = 'prefix_hex',
    topic_output_format: Literal['binary', 'prefix_hex'] = 'prefix_hex',
) -> spec.DataFrame:
    ...


@typing.overload
async def async_select_events(
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
    contract_address: spec.Address | None = None,
    event_hash: typing.Any | None = None,
    topic1: typing.Any | None = None,
    topic2: typing.Any | None = None,
    topic3: typing.Any | None = None,
    start_block: int | None = None,
    end_block: int | None = None,
    columns: typing.Sequence[str] | None = None,
    output_format: toolsql.QueryOutputFormat = 'polars',
    binary_output_format: Literal['binary', 'prefix_hex'] = 'prefix_hex',
    topic_output_format: Literal['binary', 'prefix_hex'] = 'prefix_hex',
) -> toolsql.SelectOutputData:
    ...


async def async_select_events(
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
    contract_address: spec.Address | None = None,
    event_hash: typing.Any | None = None,
    topic1: typing.Any | None = None,
    topic2: typing.Any | None = None,
    topic3: typing.Any | None = None,
    start_block: int | None = None,
    end_block: int | None = None,
    columns: typing.Sequence[str] | None = None,
    output_format: toolsql.QueryOutputFormat = 'polars',
    binary_output_format: Literal['binary', 'prefix_hex'] = 'prefix_hex',
    topic_output_format: Literal['binary', 'prefix_hex'] = 'prefix_hex',
) -> toolsql.SelectOutputData:
    # get table
    table = schema_utils.get_table_schema('events', context=context)

    # create query attribute filters
    where_equals: typing.Mapping[str, typing.Any] | None
    where_equals = {
        'contract_address': contract_address,
        'event_hash': event_hash,
        'topic1': topic1,
        'topic2': topic2,
        'topic3': topic3,
    }
    where_equals = {
        k: evm.to_binary(v)
        for k, v in where_equals.items()
        if v is not None
    }
    if len(where_equals) == 0:
        where_equals = None

    # create block range filters
    if start_block is not None:
        where_gte: typing.Mapping[str, typing.Any] | None = {
            'block_number': start_block
        }
    else:
        where_gte = None
    if end_block is not None:
        where_lte: typing.Mapping[str, typing.Any] | None = {
            'block_number': end_block
        }
    else:
        where_lte = None

    column_exprs = _get_columns(
        columns=columns,
        table=table,
        binary_output_format=binary_output_format,
        topic_output_format=topic_output_format,
    )

    # dispatch query
    result = await toolsql.async_select(
        conn=conn,
        table=table,
        where_equals=where_equals,
        where_lte=where_lte,
        where_gte=where_gte,
        columns=column_exprs,
        order_by=['block_number', 'transaction_index', 'log_index'],
        output_format=output_format,
    )

    return result


def _get_columns(
    *,
    columns: typing.Sequence[str] | None,
    table: toolsql.TableSchema,
    binary_output_format: Literal['binary', 'prefix_hex'] = 'prefix_hex',
    topic_output_format: Literal['binary', 'prefix_hex'] = 'prefix_hex',
) -> typing.Sequence[str] | toolsql.ColumnsExpression:
    if columns is None:
        columns = [column['name'] for column in table['columns']]
    column_exprs: toolsql.ColumnsExpression = columns
    # if binary_output_format != 'binary' or topic_output_format != 'binary':
    #     column_types = {
    #         column['name']: column['type'] for column in table['columns']
    #     }
    #     new_column: toolsql.ColumnExpression
    #     new_columns = []
    #     for column in columns:
    #         if column_types[column] == 'BLOB':

    #             if column.startswith('topic'):
    #                 if topic_output_format == 'prefix_hex':
    #                     new_column = {'column': column, 'encode': 'prefix_hex'}
    #                 elif topic_output_format == 'binary':
    #                     new_column = column
    #                 else:
    #                     raise Exception('unknown topic output format')

    #             else:
    #                 if binary_output_format == 'prefix_hex':
    #                     new_column = {'column': column, 'encode': 'prefix_hex'}
    #                 elif binary_output_format == 'binary':
    #                     new_column = column
    #                 else:
    #                     raise Exception('unknown binary output format')

    #         else:
    #             new_column = column
    #         new_columns.append(new_column)
    #     column_exprs = new_columns

    return column_exprs


async def async_select_event_queries(
    *,
    conn: toolsql.AsyncConnection,
    query_type: int | None,
    context: spec.Context = None,
    contract_address: spec.Address | spec.BinaryData | None = None,
    event_hash: typing.Any | None = None,
    topic1: typing.Any | None = None,
    topic2: typing.Any | None = None,
    topic3: typing.Any | None = None,
    start_block: int | None = None,
    end_block: int | None = None,
    none_means_null: bool = False,
    blocks_mean_bounds: bool = True,
) -> typing.Sequence[spec.DBEventQuery]:
    """return queries that match given parameters

    if start_block specified, return queries s.t. end_block >= input.start_block
    if end_block specified, return queries s.t. start_block <= input.end_block
    """

    # get table
    table = schema_utils.get_table_schema('event_queries', context=context)

    # encode inputs
    if contract_address is not None:
        contract_address = evm.to_binary(contract_address)
    if event_hash is not None:
        event_hash = evm.to_binary(event_hash)
    if topic1 is not None:
        topic1 = evm.to_binary(topic1)
    if topic2 is not None:
        topic2 = evm.to_binary(topic2)
    if topic3 is not None:
        topic3 = evm.to_binary(topic3)

    # create equality filters
    where_equals: typing.Mapping[str, typing.Any] | None
    where_equals = {
        'contract_address': contract_address,
        'event_hash': event_hash,
        'topic1': topic1,
        'topic2': topic2,
        'topic3': topic3,
    }
    if not none_means_null:
        where_equals = {k: v for k, v in where_equals.items() if v is not None}
    if query_type is not None:
        where_equals['query_type'] = query_type
    if len(where_equals) == 0:
        where_equals = None

    # create block number filters
    where_gte = None
    where_lte = None
    if start_block is not None:
        if blocks_mean_bounds:
            where_gte = {'end_block': start_block}
        else:
            if where_equals is None:
                where_equals = {}
            where_equals['start_block'] = start_block
    if end_block is not None:
        if blocks_mean_bounds:
            where_lte = {'start_block': end_block}
        else:
            if where_equals is None:
                where_equals = {}
            where_equals['end_block'] = end_block

    # dispatch query
    results: typing.Sequence[spec.DBEventQuery] = await toolsql.async_select(  # type: ignore
        conn=conn,
        table=table,
        where_equals=where_equals,
        where_lte=where_lte,
        where_gte=where_gte,
    )

    if results is None:
        return None
    else:
        return results


async def async_delete_event_query(
    *,
    query_id: int,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> None:
    table = schema_utils.get_table_schema('event_queries', context=context)

    await toolsql.async_delete(
        conn=conn,
        table=table,
        where_equals={'query_id': query_id},
    )


async def async_delete_event_queries(
    *,
    query_ids: typing.Sequence[int],
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> None:
    table = schema_utils.get_table_schema('event_queries', context=context)

    await toolsql.async_delete(
        conn=conn,
        table=table,
        where_in={'query_id': query_ids},
    )




from __future__ import annotations

import typing

import toolsql

from ctc import spec
from ... import schema_utils

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict

    class BlockGasRow(TypedDict):
        block_number: int
        median_gas_fee: int | float | None
        timestamp: int


async def async_upsert_median_block_gas_fee(
    block_number: int,
    *,
    median_gas_fee: int | float | None,
    timestamp: int,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> None:

    table = schema_utils.get_table_schema('block_gas', context=context)

    row = {
        'block_number': block_number,
        'median_gas_fee': median_gas_fee,
        'timestamp': timestamp,
    }
    await toolsql.async_insert(
        conn=conn,
        table=table,
        row=row,
        upsert=True,
    )


async def async_upsert_median_blocks_gas_fees(
    block_gas_data: typing.Sequence[BlockGasRow],
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> None:

    table = schema_utils.get_table_schema('block_gas', context=context)

    await toolsql.async_insert(
        conn=conn,
        table=table,
        rows=block_gas_data,
        upsert=True,
    )


async def async_select_median_block_gas_fee(
    block_number: int,
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> BlockGasRow | None:

    table = schema_utils.get_table_schema('block_gas', context=context)

    result: BlockGasRow = await toolsql.async_select(  # type: ignore
        conn=conn,
        table=table,
        where_equals={'block_number': block_number},
        output_format='single_dict_or_none',
    )
    return result


async def async_select_median_blocks_gas_fees(
    block_numbers: typing.Sequence[int],
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> typing.Mapping[int, BlockGasRow] | None:

    table = schema_utils.get_table_schema('block_gas', context=context)

    results = await toolsql.async_select(
        conn=conn,
        table=table,
        where_in={'block_number': block_numbers},
    )
    if results is None:
        return None
    else:
        return {result['block_number']: result for result in results}  # type: ignore


async def async_delete_block_gas(
    block_number: int,
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> None:

    table = schema_utils.get_table_schema('block_gas', context=context)

    await toolsql.async_delete(
        conn=conn,
        table=table,
        where_equals={'block_number': block_number},
    )


async def async_delete_blocks_gasses(
    block_numbers: typing.Sequence[int],
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> None:

    table = schema_utils.get_table_schema('block_gas', context=context)

    await toolsql.async_delete(
        conn=conn,
        table=table,
        where_in={'block_number': block_numbers},
    )



from __future__ import annotations

import typing

import tooltime

from ctc import spec


def filter_twap(
    *,
    raw_values: typing.Sequence[typing.Any],
    timestamps: typing.Sequence[typing.Any],
    filter_duration: tooltime.Timestamp,
) -> spec.DataFrame:
    """convert raw value of a TWAP"""

    import numpy as np
    import polars as pl

    # compute twap times
    timestamps_array: spec.NumpyArray = np.array(timestamps)
    filter_seconds = tooltime.timestamp_to_seconds(filter_duration)
    first_input_timestamp = timestamps[0]
    output_mask = timestamps_array > first_input_timestamp + filter_seconds
    twap_times = timestamps_array[output_mask]

    # compute twap values
    # cannot use efficient numpy operation because filter size is variable
    # (because block times are variable)
    raw_values_array: spec.NumpyArray = np.array(raw_values)
    twap_values = []
    for twap_time in twap_times:
        lower_bound = twap_time - filter_seconds < timestamps_array
        upper_bound = timestamps_array <= twap_time
        block_mask = lower_bound * upper_bound
        twap_value = raw_values_array[block_mask].mean()
        twap_values.append(twap_value)

    # format as Series
    return pl.DataFrame({'timestamp': twap_times, 'value': twap_values})




from __future__ import annotations

import asyncio
import typing

from ctc import evm
from ctc import rpc
from ctc import spec

from . import aave_oracle
from . import aave_spec


async def async_get_unclaimed_rewards(
    wallet: spec.Address,
    *,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> int:
    aave_incentives_controller = aave_spec.get_aave_address(
        'IncentivesController',
        context=context,
    )

    result = await rpc.async_eth_call(
        to_address=aave_incentives_controller,
        function_name='getUserUnclaimedRewards',
        function_parameters=[wallet],
        block_number=block,
        context=context,
    )

    if not isinstance(result, int):
        raise Exception('invalid rpc result')

    return result


async def async_get_unclaimed_rewards_by_block(
    wallet: spec.Address,
    blocks: typing.Sequence[spec.BlockNumberReference],
    *,
    context: spec.Context = None,
) -> typing.Sequence[int]:
    coroutines = [
        async_get_unclaimed_rewards(
            wallet=wallet,
            block=block,
            context=context,
        )
        for block in blocks
    ]
    return await asyncio.gather(*coroutines)


async def async_compute_wallet_rewards(
    wallet: spec.Address,
    blocks: typing.Sequence[spec.BlockNumberReference],
    *,
    context: spec.Context = None,
    replace_symbol: bool = True,
) -> typing.Mapping[str, spec.NumpyArray]:

    import numpy as np

    # add reward token
    reward_token = '0x4da27a545c0c5b758a6ba100e3a049001de870f5'
    reward_token_unstaked = '0x7fc66500c84a76ad7e9c93437bfc5ac33e2ddae9'

    reward_unclaimed_coroutine = async_get_unclaimed_rewards_by_block(
        wallet=wallet,
        blocks=blocks,
        context=context,
    )
    reward_in_wallet_coroutine = evm.async_get_erc20_balance_by_block(
        wallet=wallet,
        token=reward_token,
        blocks=blocks,
        context=context,
    )
    reward_price_coroutine = aave_oracle.async_get_asset_price_by_block(
        asset=reward_token_unstaked,
        blocks=blocks,
        context=context,
    )

    (reward_unclaimed, reward_in_wallet, reward_price,) = await asyncio.gather(
        reward_unclaimed_coroutine,
        reward_in_wallet_coroutine,
        reward_price_coroutine,
    )

    reward_unclaimed_array = np.array(reward_unclaimed) / 1e18
    reward_in_wallet_array = np.array(reward_in_wallet)
    reward_balance = reward_unclaimed_array + reward_in_wallet_array
    reward_balance_usd = reward_balance * reward_price

    if replace_symbol:
        return {
            'stkAAVE_unclaimed': reward_unclaimed_array,
            'stkAAVE_in_wallet': reward_in_wallet_array,
            'stkAAVE_balance': reward_balance,
            'stkAAVE_balance_usd': reward_balance_usd,
        }
    else:
        return {
            'reward_unclaimed': reward_unclaimed_array,
            'reward_in_wallet': reward_in_wallet_array,
            'reward_balance': reward_balance,
            'reward_balance_usd': reward_balance_usd,
        }



from __future__ import annotations

import typing

from ctc import evm
from ctc import rpc
from ctc import spec

from . import curve_spec
from . import pool_metadata
from . import pool_state


async def async_get_metapool_trade(
    metapool: spec.Address,
    *,
    token_sold: typing.Union[spec.Address, str],
    token_bought: typing.Union[spec.Address, str],
    amount_sold: typing.Union[int, float],
    input_normalized: bool = True,
    normalize_output: bool = True,
    parent_pool: typing.Optional[spec.Address] = None,
    parent_lp: typing.Optional[spec.Address] = None,
    parent_coins: typing.Optional[list[str]] = None,
    context: spec.Context = None,
) -> curve_spec.CurveTrade:

    if parent_pool is None and parent_lp is None:
        parent_pool = curve_spec.three_pool
        parent_lp = curve_spec.three_pool_lp
        parent_coins = curve_spec.three_pool_coins
    if parent_pool is None or parent_lp is None or parent_coins is None:
        raise Exception('must specify more parent parameters')

    metadata = await pool_metadata.async_get_pool_metadata(
        pool=metapool,
        context=context,
    )

    if token_bought in parent_coins:

        sold_index = await pool_metadata.async_get_token_index(
            pool=metapool, token=token_sold, metadata=metadata, context=context,
        )
        bought_index = await pool_metadata.async_get_token_index(
            pool=metapool, token=parent_lp, metadata=metadata, context=context,
        )

        if input_normalized:
            amount_sold *= 10 ** metadata['token_decimals'][sold_index]

        function_abi: spec.FunctionABI = {
            'inputs': [
                {
                    'name': 'i',
                    'type': 'int128',
                },
                {
                    'name': 'j',
                    'type': 'int128',
                },
                {
                    'name': 'dx',
                    'type': 'uint256',
                },
            ],
            'name': 'get_dy',
            'outputs': [
                {
                    'name': '',
                    'type': 'uint256',
                },
            ],
            'stateMutability': 'view',
            'type': 'function',
        }

        lp_bought = await rpc.async_eth_call(
            to_address=metapool,
            function_abi=function_abi,
            function_parameters=[sold_index, bought_index, amount_sold],
            context=context,
        )

        amount_bought = await pool_state.async_get_lp_withdrawal(
            pool=parent_pool,
            amount_lp=lp_bought,
            token_withdrawn=token_bought,
            context=context,
        )

        if normalize_output:
            # bought token can be different from lp token
            bought_decimals = await evm.async_get_erc20_decimals(
                metadata['token_addresses'][bought_index],
                context=context,
            )
            amount_bought /= 10 ** bought_decimals
            amount_sold /= 10 ** metadata['token_decimals'][sold_index]

    elif token_sold in parent_coins:

        sold_index = await pool_metadata.async_get_token_index(
            pool=metapool, token=parent_lp, metadata=metadata, context=context
        )
        bought_index = await pool_metadata.async_get_token_index(
            pool=metapool, token=token_bought, metadata=metadata, context=context
        )

        raise NotImplementedError()

    else:
        raise Exception('could not determine token indices')

    return {
        'token_sold': token_sold,
        'token_bought': token_bought,
        'amount_sold': amount_sold,
        'amount_bought': amount_bought,
    }



from __future__ import annotations

import toolsql

from ctc import config
from ctc import spec
from ... import management

from . import contract_abis_statements


async def async_intake_contract_abi(
    abi: spec.ContractABI,
    contract_address: spec.Address,
    *,
    context: spec.Context = None,
    includes_proxy: bool,
) -> None:
    """block age not checked because abis are not usually acquired from reorg-ably young blocks"""

    # validate input format
    if not isinstance(abi, list):
        raise Exception('bad format for contract abi')
    if any(not isinstance(item, dict) for item in abi):
        raise Exception('bad format for contract abi')

    if not management.get_active_schemas().get('contract_abis'):
        return

    db_config = config.get_context_db_config(
        schema_name='contract_abis',
        context=context,
    )
    async with toolsql.async_connect(db_config) as conn:
        await contract_abis_statements.async_upsert_contract_abi(
            address=contract_address,
            abi=abi,
            includes_proxy=includes_proxy,
            conn=conn,
            context=context,
        )




from __future__ import annotations

import typing

from ctc import spec

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict

    class PoolTokensMetadata(TypedDict):
        x_address: spec.Address
        y_address: spec.Address
        x_symbol: str
        y_symbol: str
        x_decimals: int
        y_decimals: int

    class PoolState(TypedDict):
        x_reserves: typing.Union[int, float]
        y_reserves: typing.Union[int, float]
        lp_total_supply: typing.Union[int, float]

    class PoolStateByBlock(TypedDict):
        x_reserves: list[typing.Union[int, float]]
        y_reserves: list[typing.Union[int, float]]
        lp_total_supply: list[typing.Union[int, float]]


uniswap_v2_factory = '0x5c69bee701ef814a2b6a3edd4b1652cb9cc5aa6f'

trade_fee = 0.003

factory_event_abis: typing.Mapping[str, spec.EventABI] = {
    'PairCreated': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'token0',
                'type': 'address',
            },
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'token1',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'pair',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': '',
                'type': 'uint256',
            },
        ],
        'name': 'PairCreated',
        'type': 'event',
    },
}


pool_function_abis: typing.Mapping[str, spec.FunctionABI] = {
    'token0': {
        'inputs': [],
        'name': 'token0',
        'outputs': [
            {
                'internalType': 'address',
                'name': '',
                'type': 'address',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'token1': {
        'inputs': [],
        'name': 'token1',
        'outputs': [
            {
                'internalType': 'address',
                'name': '',
                'type': 'address',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
}


pool_event_abis: typing.Mapping[str, spec.EventABI] = {
    'Burn': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'sender',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'amount0',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'amount1',
                'type': 'uint256',
            },
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'to',
                'type': 'address',
            },
        ],
        'name': 'Burn',
        'type': 'event',
    },
    'Mint': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'sender',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'amount0',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'amount1',
                'type': 'uint256',
            },
        ],
        'name': 'Mint',
        'type': 'event',
    },
    'Swap': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'sender',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'amount0In',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'amount1In',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'amount0Out',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'amount1Out',
                'type': 'uint256',
            },
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'to',
                'type': 'address',
            },
        ],
        'name': 'Swap',
        'type': 'event',
    },
}




import pytest

from ctc.defi import dex_utils


dexes = [
    'Balancer',
    'Curve',
    'Sushi',
    'Uniswap V2',
    'Uniswap V3',
]


@pytest.mark.parametrize('dex_name', dexes)
@pytest.mark.parametrize('network', [1])
def test_dex_directory_matches_classes(dex_name, network):

    dex = dex_utils.get_dex(dex_name)
    dex_factories = set(dex.get_pool_factories(network))

    dex_names_of_factories = dex_utils.get_dex_names_of_factories(
        network=network
    )
    directory_factories = set(
        factory
        for factory, other_dex_name in dex_names_of_factories.items()
        if other_dex_name == dex_name
    )

    assert directory_factories == dex_factories



# example_contracts = {
#     'chainlink_feed': '0x31e0a88fecB6eC0a411DBe0e9E76391498296EE9',
#     'uniswap_v2_pool': '0x9928e4046d7c6513326ccea028cd3e7a91c7590a',
#     'uniswap_v3_pool': '0x88e6a0c2ddd26feeb64f039a2c41296fcb3f5640',
#     'curve_pool': '0x06cb22615ba53e60d67bf6c341a0fd5e718e1655',
#     'sushi_pool': '0x397ff1542f962076d0bfe58ea045ffa2d347aca0',
#     'balancer_v2_pool': '0xc35bdda2e93c401c6645e0d8a0b2c86906c51710',
# }

# for target_contract_type, address in example_contracts.items():
#     print(target_contract_type)
#     detected_contract_type = await async_get_contract_type(address)
#     print(detected_contract_type)



from __future__ import annotations

import typing

from typing_extensions import Literal, TypedDict, TypeAlias

from . import binary_types
from . import transaction_types


TraceOutputType = Literal['trace', 'stateDiff', 'vmTrace']
TraceItemType = Literal['call', 'create', 'suicide', 'reward']
TraceCallType = Literal['call', 'staticcall', 'delegatecall']


TraceCallAction = TypedDict(
    'TraceCallAction',
    {
        'from': str,
        'gas': int,
        'value': int,
        #
        # call-specific keys
        'call_type': TraceCallType,
        'input': str,
        'to': str,
    },
)

TraceCreateAction = TypedDict(
    'TraceCreateAction',
    {
        'from': str,
        'gas': int,
        'value': int,
        #
        # create-specific keys
        'init': str,
    },
)


TraceSuicideAction = TypedDict(
    'TraceSuicideAction',
    {
        'from': str,
        'gas': int,
        'value': int,
        #
        # suicide-specific keys
        'init': str,
    },
)


class TraceRewardAction(TypedDict):
    value: int
    author: str
    reward_type: Literal['block', 'uncle']


TraceAction = typing.Union[
    TraceCreateAction,
    TraceCallAction,
    TraceSuicideAction,
    TraceRewardAction,
]


class TraceCallResult(TypedDict):
    gas_used: int
    output: str


class TraceCreateResult(TypedDict):
    gas_used: int
    address: str
    code: str


TraceSuicideResult: TypeAlias = None
TraceRewardResult: TypeAlias = None


TraceResult = typing.Union[
    TraceCallResult,
    TraceCreateResult,
    TraceSuicideResult,
    TraceRewardResult,
]


class SingleTrace(TypedDict):
    action: TraceAction
    block_hash: str
    block_number: int
    result: TraceResult
    subtraces: int
    trace_address: typing.Sequence[int]
    transaction_hash: str
    transaction_position: int
    type: TraceItemType


TraceList = typing.Sequence[SingleTrace]


IntegerStateDiffEqual = Literal['=']
IntegerFromTo = TypedDict('IntegerFromTo', {'from': int, 'to': int})
IntegerStateDiffMult = TypedDict('IntegerStateDiffMult', {'*': IntegerFromTo})
IntegerStateDiffAdd = TypedDict(
    'IntegerStateDiffAdd',
    {'+': int},
)
IntegerStateDiffSubtract = TypedDict(
    'IntegerStateDiffSubtract',
    {'-': int},
)
IntegerStateDiff = typing.Union[
    IntegerStateDiffMult,
    IntegerStateDiffAdd,
    IntegerStateDiffSubtract,
]


BinaryStateDiffEqual = Literal['=']
BinaryFromTo = TypedDict(
    'BinaryFromTo',
    {'from': binary_types.PrefixHexData, 'to': binary_types.PrefixHexData},
)
BinaryStateDiffMult = TypedDict('BinaryStateDiffMult', {'*': BinaryFromTo})
BinaryStateDiffAdd = TypedDict(
    'BinaryStateDiffAdd',
    {'+': binary_types.PrefixHexData},
)
BinaryStateDiffSubtract = TypedDict(
    'BinaryStateDiffSubtract',
    {'-': binary_types.PrefixHexData},
)
BinaryStateDiff = typing.Union[
    BinaryStateDiffMult,
    BinaryStateDiffAdd,
    BinaryStateDiffSubtract,
]


StorageStateDiffTrace = typing.Mapping[
    binary_types.PrefixHexData,
    BinaryStateDiff,
]


class AddressStateDiffTrace(TypedDict):
    balance: IntegerStateDiff
    code: BinaryStateDiff
    nonce: IntegerStateDiff
    storage: StorageStateDiffTrace


StateDiffTrace = typing.Mapping[
    transaction_types.TransactionHash, AddressStateDiffTrace
]


Opcode = str


class OpcodeExecutionMem(TypedDict):
    data: binary_types.PrefixHexData
    off: int


class OpcodeExecutionStore(TypedDict):
    key: binary_types.PrefixHexData
    val: binary_types.PrefixHexData


class OpcodeExecution(TypedDict):
    mem: None | OpcodeExecutionMem
    push: typing.Sequence[binary_types.PrefixHexData]
    store: None | OpcodeExecutionStore
    used: int


class OpcodeTrace(TypedDict):
    cost: int
    ex: None | OpcodeExecution
    pc: int
    sub: None | typing.Sequence['VMTrace']
    op: Opcode
    idx: str


class VMTrace(TypedDict):
    code: binary_types.PrefixHexData
    ops: typing.Sequence[OpcodeTrace]


class TraceReplayResult(TypedDict):
    trace: TraceList | None
    state_diff: StateDiffTrace | None
    vm_trace: VMTrace | None


#
# # debug traces
#


class DebugTransactionTrace(TypedDict):
    gas: int
    failed: bool
    return_value: str
    struct_logs: typing.Sequence[DebugTraceStructLog]


class DebugTraceStructLog(TypedDict):
    pc: int
    op: Opcode
    gas: int
    gas_cost: int
    depth: int
    stack: typing.Sequence[binary_types.PrefixHexData]
    memory: typing.Sequence[binary_types.RawHexData]


DebugBlockTrace = typing.Sequence[DebugTransactionTrace]




from __future__ import annotations

import toolcli

from ctc.protocols import uniswap_v2_utils
from ctc.cli import cli_utils
from ctc import spec


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_burns_command,
        'help': 'output information about pool burns',
        'args': [
            {'name': 'pool', 'help': 'pool address'},
            {'name': '--blocks', 'help': 'block number range'},
            {
                'name': '--export',
                'default': 'stdout',
                'help': 'file path for export (.json or .csv)',
            },
            {
                'name': '--overwrite',
                'action': 'store_true',
                'help': 'specify that output path can be overwritten',
            },
        ],
        'examples': [
            '0xae461ca67b15dc8dc81ce7615e0320da1a9ab8d5 --blocks 14000000:14001000',
        ],
    }


async def async_burns_command(
    *,
    pool: spec.Address,
    blocks: str | None,
    export: str,
    overwrite: bool,
) -> None:

    if blocks is not None:
        start_block, end_block = await cli_utils.async_parse_block_range(blocks)
    else:
        start_block = None
        end_block = None

    burns = await uniswap_v2_utils.async_get_pool_burns(
        pool,
        start_block=start_block,
        end_block=end_block,
    )
    cli_utils.output_data(burns, export, overwrite=overwrite)



from __future__ import annotations

import typing

import msgspec


#
# # msgspec
#


CallAction = msgspec.defstruct(
    'CallAction',
    [
        ('from', typing.Optional[str], None),  # type: ignore
        ('to', typing.Optional[str], None),  # type: ignore
        ('value', typing.Optional[str], None),  # type: ignore
        ('author', typing.Optional[str], None),  # type: ignore
        ('address', typing.Optional[str], None),  # type: ignore
        ('refundAddress', typing.Optional[str], None),  # type: ignore
        ('balance', typing.Optional[str], None),  # type: ignore
    ],
    omit_defaults=True,
)

CallResult = msgspec.defstruct(
    'CallResult',
    [
        ('address', typing.Optional[str], None),  # type: ignore
    ],
    omit_defaults=True,
)


class CallTrace(msgspec.Struct, rename='camel'):
    type: str
    action: CallAction  # type: ignore
    subtraces: int
    trace_address: typing.Sequence[int]
    result: typing.Optional[CallResult] = None  # type: ignore
    error: typing.Union[str, None] = None


class TransactionReplay(msgspec.Struct, rename='camel'):
    transaction_hash: str
    trace: list[CallTrace]


class RpcResult(msgspec.Struct):
    result: list[TransactionReplay]


if typing.TYPE_CHECKING:
    from . import create_trace_decoder

    T = typing.TypeVar('T', CallTrace, create_trace_decoder.CreateTrace)


decoder = msgspec.json.Decoder(RpcResult)


def decode_native_transfers(
    responses: typing.Sequence[str],
    block_numbers: typing.Sequence[int],
) -> typing.Sequence[typing.Any]:
    blocks_replays = [decoder.decode(response).result for response in responses]

    # transform replays into eth transfers
    transfers = []
    for block_number, block_replay in zip(block_numbers, blocks_replays):
        transfer_index = 0
        for tx_replay in block_replay:
            for trace in filter_failed_traces(tx_replay.trace):
                transfer = native_transfer_from_call_trace(
                    trace=trace,
                    block_number=block_number,
                    transfer_index=transfer_index,
                    tx_hash=tx_replay.transaction_hash,
                )
                if transfer is not None:
                    transfers.append(transfer)
                    transfer_index += 1

    return transfers


def filter_failed_traces(traces: list[T]) -> list[T]:
    keep = []
    in_error = False
    error_address = None
    len_error_address = 0
    for t, trace in enumerate(traces):
        # restart fro each transaction
        if trace.trace_address == []:
            in_error = False
            error_address = None

        # if in an error, check if next trace is still in error
        if in_error:
            if (
                len(trace.trace_address) >= len_error_address
                and trace.trace_address[:len_error_address] == error_address
            ):
                keep.append(False)
                continue
            else:
                in_error = False
                error_address = None

        # check if current trace is start of an error
        if trace.error not in [None, '', 'None']:
            in_error = True
            error_address = trace.trace_address
            len_error_address = len(error_address)
            keep.append(False)
        else:
            keep.append(True)

    return [trace for trace, keep_trace in zip(traces, keep) if keep_trace]


def native_transfer_from_call_trace(
    *, trace: CallTrace, block_number: int, transfer_index: int, tx_hash: str
) -> typing.Sequence[typing.Any] | None:
    ttype = trace.type

    # parse value
    if ttype == 'suicide':
        value = trace.action.balance  # type: ignore
    else:
        value = trace.action.value  # type: ignore
    if value == '0x0':
        return None

    # parse from address and to address
    if ttype == 'reward':
        from_address = '0x0000000000000000000000000000000000000000'
        to_address = trace.action.author  # type: ignore
    elif ttype == 'create':
        if trace.result is None:
            return None
        from_address = getattr(trace.action, 'from')
        to_address = trace.result.address
    elif ttype == 'call':
        from_address = getattr(trace.action, 'from')
        to_address = trace.action.to  # type: ignore
    elif ttype == 'suicide':
        from_address = trace.action.address  # type: ignore
        to_address = trace.action.refundAddress  # type: ignore
    else:
        raise Exception(ttype)

    return [
        block_number,
        transfer_index,
        tx_hash,
        to_address,
        from_address,
        value,
    ]




# https://docs.chain.link/docs/faq/
from __future__ import annotations

import os
import typing

from ctc import spec

if typing.TYPE_CHECKING or os.environ.get('BUILDING_SPHINX') == '1':
    from typing_extensions import TypedDict

    _FeedReference = typing.Union[spec.Address, str]

    class FeedRoundData(TypedDict):
        answer: typing.Union[int, float]
        round_id: int
        timestamp: int


path_templates = {
    'feed': '{chainlink_view}/feeds/{feed}/{feed}__{start_block}_to_{end_block}.csv',
}


feed_function_abis: dict[str, spec.FunctionABI] = {
    'aggregator': {
        'inputs': [],
        'name': 'aggregator',
        'outputs': [
            {
                'internalType': 'address',
                'name': '',
                'type': 'address',
            }
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'decimals': {
        'inputs': [],
        'name': 'decimals',
        'outputs': [{'internalType': 'uint8', 'name': '', 'type': 'uint8'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    'description': {
        'inputs': [],
        'name': 'description',
        'outputs': [{'internalType': 'string', 'name': '', 'type': 'string'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    'latestAnswer': {
        'inputs': [],
        'name': 'latestAnswer',
        'outputs': [{'internalType': 'int256', 'name': '', 'type': 'int256'}],
        'stateMutability': 'view',
        'type': 'function',
    },
    'latestRoundData': {
        'inputs': [],
        'name': 'latestRoundData',
        'outputs': [
            {'internalType': 'uint80', 'name': 'roundId', 'type': 'uint80'},
            {'internalType': 'int256', 'name': 'answer', 'type': 'int256'},
            {'internalType': 'uint256', 'name': 'startedAt', 'type': 'uint256'},
            {'internalType': 'uint256', 'name': 'updatedAt', 'type': 'uint256'},
            {
                'internalType': 'uint80',
                'name': 'answeredInRound',
                'type': 'uint80',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
}

aggregator_event_abis: dict[str, spec.EventABI] = {
    'AnswerUpdated': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'int256',
                'name': 'current',
                'type': 'int256',
            },
            {
                'indexed': True,
                'internalType': 'uint256',
                'name': 'roundId',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'updatedAt',
                'type': 'uint256',
            },
        ],
        'name': 'AnswerUpdated',
        'type': 'event',
    },
}




"""
see https://github.com/curvefi/curve-pool-registry
"""
from __future__ import annotations

import typing

from ctc import spec

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict

    class CurvePoolMetadata(TypedDict):
        token_addresses: typing.Sequence[spec.Address]
        token_symbols: typing.Sequence[str]
        token_decimals: typing.Sequence[int]
        A: int

    class CurveTrade(TypedDict):
        token_sold: spec.Address
        token_bought: spec.Address
        amount_sold: typing.Union[int, float]
        amount_bought: typing.Union[int, float]


three_pool_lp = '0x6c3f90f043a72fa612cbac8115ee7e52bde6e490'
three_pool = '0xbebc44782c7db0a1a60cb6fe97d0b483032ff1c7'
three_pool_token_index = {
    'DAI': 0,
    'USDC': 1,
    'USDT': 2,
}
three_pool_coins = ['DAI', 'USDC', 'USDT']


pool_function_abis: typing.Mapping[str, spec.FunctionABI] = {
    'A': {
        'inputs': [],
        'name': 'A',
        'outputs': [
            {
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'future_A': {
        'inputs': [],
        'name': 'future_A',
        'outputs': [
            {
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'future_A_time': {
        'inputs': [],
        'name': 'future_A_time',
        'outputs': [
            {
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'initial_A': {
        'inputs': [],
        'name': 'initial_A',
        'outputs': [
            {
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'initial_A_time': {
        'inputs': [],
        'name': 'initial_A_time',
        'outputs': [
            {
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'calc_withdraw_one_coin': {
        'inputs': [
            {
                'name': '_burn_amount',
                'type': 'uint256',
            },
            {
                'name': 'i',
                'type': 'int128',
            },
        ],
        'name': 'calc_withdraw_one_coin',
        'outputs': [
            {
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'get_dy': {
        'inputs': [
            {
                'name': 'i',
                'type': 'int128',
            },
            {
                'name': 'j',
                'type': 'int128',
            },
            {
                'name': 'dx',
                'type': 'uint256',
            },
        ],
        'name': 'get_dy',
        'outputs': [
            {
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
    'get_virtual_price': {
        'inputs': [],
        'name': 'get_virtual_price',
        'outputs': [
            {
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    },
}

pool_event_abis: typing.Mapping[str, spec.EventABI] = {
    'TokenExchange': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'name': 'buyer',
                'type': 'address',
            },
            {
                'indexed': False,
                'name': 'sold_id',
                'type': 'int128',
            },
            {
                'indexed': False,
                'name': 'tokens_sold',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'name': 'bought_id',
                'type': 'int128',
            },
            {
                'indexed': False,
                'name': 'tokens_bought',
                'type': 'uint256',
            },
        ],
        'name': 'TokenExchange',
        'type': 'event',
    },
    'TokenExchangeUnderlying': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'name': 'buyer',
                'type': 'address',
            },
            {
                'indexed': False,
                'name': 'sold_id',
                'type': 'int128',
            },
            {
                'indexed': False,
                'name': 'tokens_sold',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'name': 'bought_id',
                'type': 'int128',
            },
            {
                'indexed': False,
                'name': 'tokens_bought',
                'type': 'uint256',
            },
        ],
        'name': 'TokenExchangeUnderlying',
        'type': 'event',
    },
}




from __future__ import annotations

import os
import typing

from . import context_utils
from . import config_spec


def _add_config_env_vars(
    config: typing.MutableMapping[str, typing.Any]
) -> typing.MutableMapping[str, typing.Any]:
    config = dict(config)
    parse_env_network_and_provider(config)
    parse_env_cache(config)
    return config


def parse_env_network_and_provider(
    config: typing.MutableMapping[str, typing.Any]
) -> None:
    """
    Possible actions
    - set existing provider as default
    - create new provider and use it as default
    """

    # parse env network
    env_network = os.environ.get(config_spec.network_env_var)
    env_chain_id: None | int = None
    if env_network is not None:
        if env_network.isnumeric():
            env_chain_id = int(env_network)
        else:
            for chain_id, network_metadata in config['networks'].items():
                if network_metadata['name'] == env_network:
                    env_chain_id = chain_id
                    break
            else:
                raise Exception('unknown network: ' + str(env_network))

    # parse env provider
    env_provider_name: None | str = None
    env_provider_chain_id: None | int = None
    env_provider = os.environ.get(config_spec.provider_env_var)
    if env_provider is not None and env_provider != '':
        if env_provider in config['providers']:
            env_provider_name = env_provider
            env_provider_chain_id = config['providers'][env_provider]['network']
        else:
            for provider_name, provider in config['providers'].items():
                # if env_provider already in config, use it
                if provider['url'] == env_provider:
                    env_provider_name = provider_name
                    env_provider_chain_id = provider['network']
                    break
            else:
                # if env_provider not in config, make new provider
                from ctc import rpc

                if env_provider.startswith('{'):
                    import ast

                    provider_shorthand = ast.literal_eval(env_provider)
                else:
                    provider_shorthand = {'url': env_provider}

                try:
                    provider = rpc.resolve_provider(
                        provider_shorthand,
                        other_providers=config.get('providers'),
                    )
                    env_provider_name = provider['name']
                    env_provider_chain_id = provider['network']
                    config['providers'][env_provider_name] = provider

                except LookupError:
                    import warnings

                    warnings.warn(
                        'invalid provider specified in CTC_PROVIDER environment variable, skipping'
                    )

    # set default provider
    if env_chain_id is not None and env_provider_chain_id is not None:
        if env_chain_id == env_provider_chain_id:
            config['default_network'] = env_chain_id
        else:
            raise Exception(
                'network of CTC_NETWORK does not match network of CTC_PROVIDER'
            )
    elif env_chain_id is not None:
        config['default_network'] = env_chain_id
    elif env_provider_chain_id is not None:
        config['default_network'] = env_provider_chain_id

    # set provider
    if env_provider_name is not None:
        config['default_providers'][env_provider_chain_id] = env_provider_name


def parse_env_cache(config: typing.MutableMapping[str, typing.Any]) -> None:

    env_cache = os.environ.get(config_spec.cache_env_var)
    if env_cache is not None and env_cache != '':
        import ast

        cache_shorthand = ast.literal_eval(env_cache)
        rules = context_utils.context_caches._extract_context_cache_rules(
            context_cache=cache_shorthand
        )
        config['context_cache_rules'] = rules + config['context_cache_rules']




from __future__ import annotations

import asyncio
import functools
import typing

from ctc import evm
from ctc import rpc
from ctc import spec

from . import chainlink_db
from . import chainlink_feed_metadata


async def async_get_aggregator_description(
    aggregator: spec.Address,
    *,
    context: spec.Context = None,
) -> str:

    function_abi: spec.FunctionABI = {
        'inputs': [],
        'name': 'description',
        'outputs': [{'internalType': 'string', 'name': '', 'type': 'string'}],
        'stateMutability': 'view',
        'type': 'function',
    }

    result = await rpc.async_eth_call(
        to_address=aggregator,
        function_abi=function_abi,
        context=context,
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')
    return result


async def async_get_aggregator_base_quote(
    aggregator: spec.Address,
    *,
    context: spec.Context = None,
) -> dict[str, str]:
    description = await async_get_aggregator_description(
        aggregator=aggregator,
        context=context,
    )
    try:
        base, quote = description.split(' / ')
        return {'base': base, 'quote': quote}
    except Exception:
        raise Exception(
            'could not determine base or quote from description: '
            + str(description)
        )


async def async_get_feed_aggregator_history(
    feed: spec.Address,
    *,
    context: spec.Context = None,
) -> typing.Mapping[spec.Address, int]:
    # TODO: can make it go 2x as fast by searching at start and end concurrently
    # TODO: be able to give block range to search within

    feed = await chainlink_feed_metadata.async_resolve_feed_address(
        feed,
        context=context,
    )

    # query current feed aggregator
    aggregator_coroutine = chainlink_feed_metadata.async_get_feed_aggregator(
        feed=feed,
        block='latest',
        context=context,
    )

    # get known updates
    db_results = await chainlink_db.async_query_aggregator_updates(
        feed=feed,
        context=context,
    )
    if db_results is not None:
        db_results = sorted(
            db_results, key=lambda result: result['block_number']
        )
        feed_aggregator_updates = {
            update['aggregator']: update['block_number']
            for update in db_results
        }
    else:
        feed_aggregator_updates = {}

    # await current aggregator
    aggregator = await aggregator_coroutine

    # if current aggregator is in database, then data are up to date
    if aggregator in feed_aggregator_updates:
        return feed_aggregator_updates

    # otherwise, need to update data
    else:

        # collect new update data
        previous_aggregators = await async_get_feed_previous_aggregators(
            feed=feed,
            context=context,
        )
        previous_aggregators = [
            aggregator
            for aggregator in previous_aggregators
            if aggregator not in feed_aggregator_updates
        ]
        aggregator_start_blocks = await _async_get_aggregator_start_blocks(
            previous_aggregators=previous_aggregators,
            feed=feed,
            context=context,
        )

        # intake new data to database
        updates: typing.Sequence[
            chainlink_db.chainlink_schema_defs._FeedAggregatorUpdate
        ] = [
            {
                'feed': feed,
                'aggregator': aggregator,
                'block_number': block_number,
            }
            for aggregator, block_number in aggregator_start_blocks.items()
        ]
        await chainlink_db.async_intake_aggregator_updates(
            updates=updates,
            context=context,
        )

        return dict(feed_aggregator_updates, **aggregator_start_blocks)


async def _async_get_aggregator_start_blocks(
    previous_aggregators: typing.Sequence[spec.Address],
    feed: spec.Address,
    *,
    context: spec.Context = None,
) -> typing.Mapping[spec.Address, int]:

    print('locating Chainlink aggregator update blocks...')

    from ctc.toolbox import search_utils

    latest = await evm.async_get_latest_block_number(context=context)

    # TODO: use chainlink feed registry for this
    feed_creation_block = await evm.async_get_contract_creation_block(
        contract_address=feed,
        context=context,
    )

    aggregator_start_blocks = {}
    aggregator_start_blocks[previous_aggregators[0]] = feed_creation_block

    last_known_block = latest
    for next_aggregator in reversed(previous_aggregators[1:]):

        async_is_match = functools.partial(
            _async_aggregator_transition,
            feed=feed,
            next_aggregator=next_aggregator,
            context=context,
        )

        block = await search_utils.async_binary_search(
            async_is_match=async_is_match,
            start_index=feed_creation_block,
            end_index=last_known_block,
        )

        if block is None:
            raise Exception('could not determine start block of aggregator')

        aggregator_start_blocks[next_aggregator] = block
        last_known_block = block - 1

    return dict(
        sorted(aggregator_start_blocks.items(), key=lambda item: item[1])  # type: ignore
    )


async def _async_aggregator_transition(
    block: int,
    next_aggregator: spec.Address,
    *,
    feed: spec.Address,
    context: spec.Context = None,
) -> bool:

    aggregator = await chainlink_feed_metadata.async_get_feed_aggregator(
        feed,
        block=block,
        context=context,
    )
    return next_aggregator == aggregator


async def async_get_feed_previous_aggregators(
    feed: str,
    *,
    context: spec.Context = None,
) -> list[spec.Address]:

    feed = await chainlink_feed_metadata.async_resolve_feed_address(
        feed, context=context
    )

    phase_id_abi: spec.FunctionABI = {
        'inputs': [],
        'name': 'phaseId',
        'outputs': [{'internalType': 'uint16', 'name': '', 'type': 'uint16'}],
        'stateMutability': 'view',
        'type': 'function',
    }

    current_phase = await rpc.async_eth_call(
        to_address=feed,
        function_abi=phase_id_abi,
        context=context,
    )

    phase_aggregators_abi: spec.FunctionABI = {
        'inputs': [{'internalType': 'uint16', 'name': '', 'type': 'uint16'}],
        'name': 'phaseAggregators',
        'outputs': [
            {
                'internalType': 'contract AggregatorV2V3Interface',
                'name': '',
                'type': 'address',
            }
        ],
        'stateMutability': 'view',
        'type': 'function',
    }

    coroutines = [
        rpc.async_eth_call(
            to_address=feed,
            function_abi=phase_aggregators_abi,
            function_parameters=[i],
            context=context,
        )
        for i in range(1, current_phase + 1)
    ]
    aggregators = await asyncio.gather(*coroutines)

    return aggregators




from __future__ import annotations

import typing

from ctc.protocols import uniswap_v2_utils
from ctc import spec

if typing.TYPE_CHECKING:
    import tooltime

    from typing_extensions import Literal


async def async_get_pool_trades(
    pool_address: spec.Address,
    *,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    label: Literal['index', 'symbol', 'address'] = 'index',
    normalize: bool = True,
    context: spec.Context = None,
) -> spec.DataFrame:

    return await uniswap_v2_utils.async_get_pool_swaps(
        pool=pool_address,
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        label=label,
        normalize=normalize,
        context=context,
    )



from __future__ import annotations

import typing

from ctc import spec


async def async_is_contract_address(
    address: spec.Address,
    *,
    block: spec.BlockNumberReference = 'latest',
    context: spec.Context = None,
) -> bool:
    """return whether address has bytecode on chain"""

    from ctc import rpc

    code = await rpc.async_eth_get_code(
        address=address,
        block_number=block,
        context=context,
    )
    return len(code) >= 3


async def async_are_contract_addresses(
    addresses: typing.Sequence[spec.Address],
    *,
    block: spec.BlockNumberReference = 'latest',
    context: spec.Context = None,
) -> dict[spec.Address, bool]:
    """return whether addresses have bytecode on chain"""

    from ctc import rpc

    codes = await rpc.async_batch_eth_get_code(
        addresses=addresses,
        block_number=block,
        context=context,
    )
    return {address: len(code) > 3 for address, code in zip(addresses, codes)}



from __future__ import annotations

import typing

from ctc import spec
from .. import block_utils
from . import transaction_convert


async def async_get_transaction(
    transaction_hash: str,
    *,
    context: spec.Context = None,
) -> spec.DBTransaction:
    """get transaction"""

    import asyncio
    from ctc import config
    from ctc import db
    from ctc import rpc

    # get from database
    read_cache, write_cache = config.get_context_cache_read_write(
        context=context, schema_name='transactions'
    )
    if read_cache:
        db_tx = await db.async_query_transaction(
            hash=transaction_hash,
            context=context,
        )
        if db_tx is not None:
            return db_tx

    # get from node
    raw_tx, raw_receipt = await asyncio.gather(
        rpc.async_eth_get_transaction_by_hash(
            transaction_hash,
            context=context,
        ),
        rpc.async_eth_get_transaction_receipt(
            transaction_hash,
            context=context,
        ),
    )

    db_transaction = (
        transaction_convert.convert_rpc_transaction_to_db_transaction(
            transaction=raw_tx, receipt=raw_receipt
        )
    )

    # remove fields
    if write_cache:
        await db.async_intake_transaction(
            transaction=db_transaction,
            context=context,
            already_converted=True,
        )

    return db_transaction


async def async_get_transactions(
    transaction_hashes: typing.Sequence[str],
    *,
    context: spec.Context = None,
) -> list[spec.DBTransaction]:
    """get transactions"""

    import asyncio

    coroutines = [
        async_get_transaction(transaction_hash, context=context)
        for transaction_hash in transaction_hashes
    ]
    return await asyncio.gather(*coroutines)


#
# # block transactions
#


async def async_get_block_transactions(
    block: int,
    *,
    context: spec.Context = None,
) -> typing.Sequence[spec.DBTransaction]:
    """get all transactions in a block"""
    return await async_get_blocks_transactions(blocks=[block], context=context)


async def async_get_blocks_transactions(
    blocks: typing.Sequence[spec.BlockReference] | None = None,
    *,
    start_block: spec.BlockReference | None = None,
    end_block: spec.BlockReference | None = None,
    context: spec.Context = None,
) -> typing.Sequence[spec.DBTransaction]:
    """get all transactions in collection of blocks"""

    import ctc.config
    import ctc.db

    if blocks is not None and (
        start_block is not None or end_block is not None
    ):
        raise Exception('cannot specify both blocks and start_block/end_block')
    if blocks is None and (start_block is None or end_block is None):
        raise Exception('must specify blocks or start_block+end_block')

    if blocks is not None:
        block_numbers = await block_utils.async_block_references_to_int(
            blocks=blocks, context=context
        )
    else:
        block_numbers = None

    # get from db
    read_cache, write_cache = ctc.config.get_context_cache_read_write(
        context=context,
        schema_name='transactions',
    )
    db_txs: typing.Sequence[spec.DBTransaction] = []
    if read_cache:

        result = await ctc.db.async_query_blocks_transactions(
            block_numbers=block_numbers,
            start_block=start_block,
            end_block=end_block,
            context=context,
        )
        if result is not None:
            db_txs, manifest = result
            missing_blocks = []

            if block_numbers is None:
                if start_block is None or end_block is None:
                    raise Exception('must specify more information about block range')
                start_block = await block_utils.async_block_number_to_int(
                    start_block, context=context
                )
                end_block = await block_utils.async_block_number_to_int(
                    end_block, context=context
                )
                block_numbers = list(range(start_block, end_block + 1))

            for block_number in block_numbers:
                if block_number not in manifest:
                    missing_blocks.append(block_number)
            if len(missing_blocks) == 0:
                return db_txs
            else:
                block_numbers = missing_blocks

    from ctc import rpc

    # get blocks from node
    rpc_blocks = await rpc.async_batch_eth_get_block_by_number(
        block_numbers=block_numbers,
        include_full_transactions=True,
        context=context,
    )

    # convert rpc txs to db txs
    rpc_transactions = [
        tx for rpc_block in rpc_blocks for tx in rpc_block['transactions']
    ]
    tx_hashes = [tx['hash'] for tx in rpc_transactions]
    rpc_receipts = await rpc.async_batch_eth_get_transaction_receipt(
        transaction_hashes=tx_hashes,
        context=context,
    )
    new_db_txs = [
        transaction_convert.convert_rpc_transaction_to_db_transaction(
            transaction=rpc_transaction, receipt=rpc_receipt
        )
        for rpc_transaction, rpc_receipt in zip(rpc_transactions, rpc_receipts)
    ]
    new_db_txs.extend(db_txs)

    if write_cache:
        await ctc.db.async_intake_blocks(
            rpc_blocks=rpc_blocks,
            context=context,
            blocks_db_transactions=new_db_txs,
        )

    return new_db_txs


#
# # transaction logs
#


async def async_get_transaction_logs(
    transaction_hash: str,
    *,
    context: spec.Context = None,
) -> typing.Sequence[spec.RawLog]:
    """get raw logs emitted by a transaction"""

    from ctc import rpc

    receipt: spec.RPCTransactionReceipt = (
        await rpc.async_eth_get_transaction_receipt(
            transaction_hash, context=context
        )
    )

    return receipt['logs']


async def async_get_transactions_logs(
    transaction_hashes: typing.Sequence[str],
    *,
    context: spec.Context = None,
) -> typing.Sequence[spec.RawLog]:
    """get raw logs emitted by transactions"""

    from ctc import rpc

    receipts = await rpc.async_batch_eth_get_transaction_receipt(
        transaction_hashes=transaction_hashes, context=context
    )

    return [log for receipt in receipts for log in receipt['logs']]


#
# # transactions of an address
#


async def async_get_transaction_count(
    address: spec.Address, *, context: spec.Context = None
) -> int:
    """get transaction count of address"""

    from ctc import rpc

    result = await rpc.async_eth_get_transaction_count(address, context=context)
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result




import toolsql

import ctc
from ctc import db

import conftest


example_data = [
    {
        'number': 100,
        'difficulty': 17916437174,
        'extra_data': '0x476574682f4c5649562f76312e302e302f6c696e75782f676f312e342e32',
        'gas_limit': 5000,
        'gas_used': 0,
        'hash': '0xdfe2e70d6c116a541101cecbb256d7402d62125f6ddc9b607d49edc989825c64',
        'logs_bloom': '0x0',
        'miner': '0xbb7b8287f3f0a933474a79eae42cbca977791171',
        'mix_hash': '0x5bb43c0772e58084b221c8e0c859a45950c103c712c5b8f11d9566ee078a4501',
        'nonce': '0x37129c7f29a9364b',
        'parent_hash': '0xdb10afd3efa45327eb284c83cc925bd9bd7966aea53067c1eebe0724d124ec1e',
        'receipts_root': '0x56e81f171bcc55a6ff8345e692c0f86e5b48e01b996cadc001622fb5e363b421',
        'sha3_uncles': '0x1dcc4de8dec75d7aab85b567b6ccd41ad312451b948a7413f0a142fd40d49347',
        'size': 542,
        'state_root': '0x90c25f6d7fddeb31a6cc5668a6bba77adbadec705eb7aa5a51265c2d1e3bb7ac',
        'timestamp': 1438270443,
        'total_difficulty': '1766758139014',
        'transactions': [],
        'transactions_root': '0x56e81f171bcc55a6ff8345e692c0f86e5b48e01b996cadc001622fb5e363b421',
        'uncles': [],
    },
    {
        'number': 101,
        'difficulty': 17925185434,
        'extra_data': '0x476574682f4c5649562f76312e302e302f6c696e75782f676f312e342e32',
        'gas_limit': 5000,
        'gas_used': 0,
        'hash': '0x4f66fd0241681ebbc119f97e952c1036b87b6e8f64f5c5d84c5c7a9bb1ebfdcc',
        'logs_bloom': '0x0',
        'miner': '0xbb7b8287f3f0a933474a79eae42cbca977791171',
        'mix_hash': '0xf086f1bc68d5a15cd06ac2017e82e8a02399d6667896a224b65a32600425a08d',
        'nonce': '0x4587d7ffc0496ea5',
        'parent_hash': '0xdfe2e70d6c116a541101cecbb256d7402d62125f6ddc9b607d49edc989825c64',
        'receipts_root': '0x56e81f171bcc55a6ff8345e692c0f86e5b48e01b996cadc001622fb5e363b421',
        'sha3_uncles': '0x1dcc4de8dec75d7aab85b567b6ccd41ad312451b948a7413f0a142fd40d49347',
        'size': 542,
        'state_root': '0xc66500dc36b0af946808ab610894d1b8af50a6033e4d92ba93236edc66649c01',
        'timestamp': 1438270445,
        'total_difficulty': '1784683324448',
        'transactions': [],
        'transactions_root': '0x56e81f171bcc55a6ff8345e692c0f86e5b48e01b996cadc001622fb5e363b421',
        'uncles': [],
    },
    {
        'number': 102,
        'difficulty': 17933937965,
        'extra_data': '0x476574682f76312e302e302f6c696e75782f676f312e342e32',
        'gas_limit': 5000,
        'gas_used': 0,
        'hash': '0x16110f3aa1895de2ec22cfd746751f724d112a953c71b62858a1523b50f3dc64',
        'logs_bloom': '0x0',
        'miner': '0x28921e4e2c9d84f4c0f0c0ceb991f45751a0fe93',
        'mix_hash': '0xf3277624282cb1e4f3db0e97fb2be539674fc1d6b9c3cc91014ececa73083ba8',
        'nonce': '0x8d36eecca27a8e00',
        'parent_hash': '0x4f66fd0241681ebbc119f97e952c1036b87b6e8f64f5c5d84c5c7a9bb1ebfdcc',
        'receipts_root': '0x56e81f171bcc55a6ff8345e692c0f86e5b48e01b996cadc001622fb5e363b421',
        'sha3_uncles': '0x1dcc4de8dec75d7aab85b567b6ccd41ad312451b948a7413f0a142fd40d49347',
        'size': 537,
        'state_root': '0x02607447cef691e3d873cf5a1d88ac201813353147ee0d631789ec757accfa6b',
        'timestamp': 1438270447,
        'total_difficulty': '1802617262413',
        'transactions': [],
        'transactions_root': '0x56e81f171bcc55a6ff8345e692c0f86e5b48e01b996cadc001622fb5e363b421',
        'uncles': [],
    },
]

example_data = [
    ctc.convert_rpc_block_to_db_block(block) for block in example_data
]


async def test_blocks_crud():

    db_config = conftest.get_test_db_config()
    db_schema = db.get_prepared_schema(
        schema_name='blocks',
        context=dict(network='ethereum'),
    )
    toolsql.create_db(
        db_config=db_config,
        db_schema=db_schema,
        if_not_exists=True,
        confirm=True,
    )

    network = 1

    # insert data
    async with toolsql.async_connect(db_config) as conn:
        for block in example_data:
            await db.async_upsert_block(
                conn=conn, block=block, context=dict(network=network)
            )

    # get data individually
    async with toolsql.async_connect(db_config) as conn:
        for block in example_data:
            db_block = await db.async_select_block(
                conn=conn,
                block_number=block['number'],
                context=dict(network=network),
            )
            for key, target_value in block.items():
                assert target_value == db_block[key]

    # get data collectively
    async with toolsql.async_connect(db_config) as conn:
        block_numbers = [block['number'] for block in example_data]
        db_blocks = await db.async_select_blocks(
            conn=conn,
            block_numbers=block_numbers,
            context=dict(network=network),
        )
        db_blocks = sorted(db_blocks, key=lambda block: block['number'])
        assert db_blocks == example_data

    # delete entries one by one
    async with toolsql.async_connect(db_config) as conn:
        for block in example_data:
            await db.async_delete_block(
                conn=conn,
                block_number=block['number'],
                context=dict(network=network),
            )

    # ensure all entries deleted
    async with toolsql.async_connect(db_config) as conn:
        db_blocks = await db.async_select_blocks(
            conn=conn,
            block_numbers=block_numbers,
            context=dict(network=network),
        )
        assert all(item is None for item in db_blocks)




from __future__ import annotations

import typing
from typing_extensions import TypedDict

from . import db_types
from . import rpc_types
from . import network_types


# Context can be any of the context types, either short-hand or full
Context = typing.Union[None, int, str, 'ShorthandContext', 'NormalizedContext']


# ShorthandContext is maximally flexible, allowing many nested data types
class ShorthandContext(TypedDict, total=False):
    network: network_types.NetworkReference | None
    provider: rpc_types.ProviderReference | None
    cache: ContextCacheShorthand


# NormalizedContext is a normalized context dict with all fields present
class NormalizedContext(TypedDict):
    network: network_types.NetworkReference | None
    provider: rpc_types.ProviderReference | None
    cache: typing.Sequence[ContextCacheRule]


# ContextCacheRule specifies cache rules for contexts that satisfy filter
class ContextCacheRule(TypedDict, total=False):
    backend: str
    read: bool
    write: bool
    filter: 'ContextCacheFilter'


# ContextCacheFilter determines the scope of when rule is applicable
class ContextCacheFilter(TypedDict, total=False):
    chain_id: network_types.ChainId | None
    schema: db_types.SchemaName
    backend: str


# ContextCacheShorthand is a maximally flexible cache specification
ContextCacheShorthand = typing.Union[
    None,
    bool,
    str,
    'ContextCacheRule',
    typing.Mapping[
        typing.Union[str, int],
        typing.Union[None, bool, str, 'ContextCacheRule'],
    ],
    typing.Sequence['ContextCacheRule'],
]




from __future__ import annotations

import typing

import toolsql

from ctc import spec
from ... import management
from . import block_timestamps_statements
from ..blocks import blocks_statements


async def async_select_block_timestamp(
    block_number: int,
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> int | None:

    timestamp_schema = management.get_active_timestamp_schema()

    if timestamp_schema == 'block_timestamps':
        return await block_timestamps_statements.async_select_block_timestamp(
            conn=conn,
            block_number=block_number,
            context=context,
        )

    elif timestamp_schema == 'blocks':
        return await blocks_statements.async_select_block_timestamp(
            conn=conn,
            block_number=block_number,
            context=context,
        )

    else:
        raise Exception('unknown schema: ' + str(timestamp_schema))


async def async_select_block_timestamps(
    block_numbers: typing.Sequence[typing.SupportsInt],
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> list[int | None] | None:

    timestamp_schema = management.get_active_timestamp_schema()

    if timestamp_schema == 'block_timestamps':
        return await block_timestamps_statements.async_select_block_timestamps(
            conn=conn,
            block_numbers=block_numbers,
            context=context,
        )

    elif timestamp_schema == 'blocks':
        return await blocks_statements.async_select_block_timestamps(
            conn=conn,
            block_numbers=block_numbers,
            context=context,
        )

    else:
        raise Exception('unknown schema: ' + str(timestamp_schema))


async def async_select_max_block_number(
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> int | None:

    timestamp_schema = management.get_active_timestamp_schema()

    if timestamp_schema == 'block_timestamps':
        return await block_timestamps_statements.async_select_max_block_number(
            conn=conn,
            context=context,
        )

    elif timestamp_schema == 'blocks':
        return await blocks_statements.async_select_max_block_number(
            conn=conn,
            context=context,
        )

    else:
        raise Exception('unknown schema: ' + str(timestamp_schema))


async def async_select_max_block_timestamp(
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> int | None:
    timestamp_schema = management.get_active_timestamp_schema()

    if timestamp_schema == 'block_timestamps':
        return (
            await block_timestamps_statements.async_select_max_block_timestamp(
                conn=conn,
                context=context,
            )
        )

    elif timestamp_schema == 'blocks':
        return await blocks_statements.async_select_max_block_timestamp(
            conn=conn,
            context=context,
        )

    else:
        raise Exception('unknown schema: ' + str(timestamp_schema))

